# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# CSV Format: timestamp; type; transaction_id; address; fee; total

import logging
from csv import reader
from typing import List

from rp2.logger import create_logger
from rp2.rp2_decimal import RP2Decimal, ZERO

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.dali_configuration import Keyword
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction

_SENT: str = "SENT"
_RECV: str = "RECV"


class InputPlugin(AbstractInputPlugin):

    __TREZOR: str = "Trezor"

    __TIMESTAMP_INDEX: int = 0
    __TYPE_INDEX: int = 1
    __TRANSACTION_ID_INDEX: int = 2
    __FEE_INDEX: int = 4
    __TOTAL_INDEX: int = 5

    __DELIMITER = ";"

    def __init__(
        self,
        account_holder: str,
        account_nickname: str,
        currency: str,
        timezone: str,
        csv_file: str,
    ) -> None:

        super().__init__(account_holder)
        self.__account_nickname: str = account_nickname
        self.__currency: str = currency
        self.__timezone: str = timezone
        self.__csv_file: str = csv_file

        self.__logger: logging.Logger = create_logger(f"{self.__TREZOR}/{currency}/{self.__account_nickname}/{self.account_holder}")

    def load(self) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        with open(self.__csv_file, mode="r", encoding="utf-8") as csv_file:
            lines = reader(csv_file, delimiter=self.__DELIMITER)
            header_found: bool = False
            for line in lines:
                raw_data: str = self.__DELIMITER.join(line)
                if not header_found:
                    # Skip header line
                    header_found = True
                    self.__logger.debug("Header: %s", raw_data)
                    continue
                self.__logger.debug("Transaction: %s", raw_data)
                transaction_type: str = line[self.__TYPE_INDEX]
                spot_price: str = Keyword.UNKNOWN.value
                crypto_hash: str = line[self.__TRANSACTION_ID_INDEX]
                fee_number: RP2Decimal = RP2Decimal(line[self.__FEE_INDEX])
                total_number: RP2Decimal = RP2Decimal(line[self.__TOTAL_INDEX])

                if total_number == ZERO and fee_number > ZERO:
                    # Cost-only transaction
                    result.append(
                        OutTransaction(
                            self.__TREZOR,
                            crypto_hash,
                            raw_data,
                            f"{line[self.__TIMESTAMP_INDEX]} {self.__timezone}",
                            self.__currency,
                            self.__account_nickname,
                            self.account_holder,
                            "Sell",
                            spot_price,
                            line[self.__FEE_INDEX],
                            "0",
                            notes="Cost-only transaction",
                        )
                    )
                else:
                    result.append(
                        IntraTransaction(
                            self.__TREZOR,
                            crypto_hash,
                            raw_data,
                            f"{line[self.__TIMESTAMP_INDEX]} {self.__timezone}",
                            self.__currency,
                            self.__account_nickname if transaction_type == _SENT else Keyword.UNKNOWN.value,
                            self.account_holder if transaction_type == _SENT else Keyword.UNKNOWN.value,
                            self.__account_nickname if transaction_type == _RECV else Keyword.UNKNOWN.value,
                            self.account_holder if transaction_type == _RECV else Keyword.UNKNOWN.value,
                            spot_price,
                            str(total_number + fee_number) if transaction_type == _SENT else Keyword.UNKNOWN.value,
                            str(total_number) if transaction_type == _RECV else Keyword.UNKNOWN.value,
                        )
                    )

        return result



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argparse import ArgumentParser, Namespace
from difflib import unified_diff
from itertools import zip_longest
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, List

import ezodf

from rp2.rp2_decimal import CRYPTO_DECIMALS

_PURGE_HYPERLIKS = True


def _row_as_string(row: Any) -> str:
    values: List[str] = []
    cell: Any
    for cell in row:
        value: Any = str(_parse_cell_value(cell))
        values.append(value)

    # Remove trailing whitespace from row
    i: int
    for i in range(len(values) - 1, -1, -1):
        if not values[i]:
            del values[i]
        else:
            break

    return ",".join(values)


def _parse_cell_value(cell: Any) -> Any:
    value: Any
    try:
        if cell.formula:
            value = cell.formula
            if _PURGE_HYPERLIKS and value.startswith("=HYPERLINK"):
                value = value.split(";")[1].lstrip(' "').rstrip('")')
        elif cell.value or cell.value == 0:
            value = round(float(cell.value), CRYPTO_DECIMALS)
            if value == -0.0:
                value = 0
        else:
            value = ""
    except ValueError:
        value = cell.value

    return value


def ods_diff(file1_path: Path, file2_path: Path, generate_ascii_representation: bool) -> str:  # pylint: disable=too-many-branches

    if not file1_path.exists():
        return f"Error: {file1_path} does not exist"
    if not file2_path.exists():
        return f"Error: {file2_path} does not exist"

    file1: Any = ezodf.opendoc(str(file1_path))
    file2: Any = ezodf.opendoc(str(file2_path))
    sheet1: Any
    sheet2: Any

    contents1: List[str] = []
    contents2: List[str] = []
    row_count1: int = 0
    row_count2: int = 0
    for sheet1 in file1.sheets:
        if sheet1.name not in file2.sheets.names():
            contents2.append(f"{sheet1.name}: sheet not found in '{file2_path}'")
            continue
        sheet2 = file2.sheets[sheet1.name]
        contents1.append(sheet1.name)
        contents2.append(sheet1.name)

        row1: Any = None
        row2: Any = None
        for _, (row1, row2) in enumerate(zip_longest(sheet1.rows(), sheet2.rows())):
            string_row: str
            if row1:
                string_row = _row_as_string(row1)
                if string_row:
                    contents1.append(string_row)
                    row_count1 += 1
            if row2:
                string_row = _row_as_string(row2)
                if string_row:
                    contents2.append(string_row)
                    row_count2 += 1

    if row_count1 <= 0 or row_count2 <= 0:
        return f"Error: {file1_path} has no data in common with {file2_path}"

    for sheet2 in file2.sheets:
        if sheet2.name not in file1.sheets.names():
            contents1.append(f"{sheet2.name}: sheet not found in '{file1_path}'")

    if generate_ascii_representation:
        for file_path, contents in zip([file1_path, file2_path], [contents1, contents2]):
            with NamedTemporaryFile(mode="w", delete=False) as temp_file:
                temp_file.write("\n".join(contents))
                temp_file.flush()
                print(f"ASCII representation of {file_path}: {temp_file.name}")

    return "\n".join(unified_diff(contents1, contents2, lineterm=""))


def main() -> None:

    parser: ArgumentParser = ArgumentParser(description="Generate yearly capital gain/loss report and account balances for crypto holdings.")
    parser.add_argument(
        "-a",
        "--ascii-representation",
        action="store_true",
        help="Save ASCII representation of ODS files in temporary directory",
    )
    parser.add_argument(
        "ods_file1",
        action="store",
        help="First ODS file",
        metavar="ODS_FILE1",
        type=str,
    )
    parser.add_argument(
        "ods_file2",
        action="store",
        help="Second ODS file",
        metavar="ODS_FILE2",
        type=str,
    )

    args: Namespace = parser.parse_args()
    print(ods_diff(Path(args.ods_file1), Path(args.ods_file2), generate_ascii_representation=args.ascii_representation))


if __name__ == "__main__":
    main()



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type

import ezodf
from rp2.rp2_error import RP2TypeError

from dali.abstract_transaction import AbstractTransaction
from dali.dali_configuration import Keyword, is_crypto_field, is_fiat_field, is_internal_field
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.logger import LOGGER
from dali.out_transaction import OutTransaction

_MIN_ROWS: int = 40
_MAX_COLUMNS: int = 20
_TABLE_END: str = "TABLE END"

_IN: str = "IN"
_OUT: str = "OUT"
_INTRA: str = "INTRA"

_TABLE_ORDER: Dict[str, int] = {
    _IN: 1,
    _OUT: 2,
    _INTRA: 3,
}

_TABLE_TO_SECTION_NAME: Dict[str, Keyword] = {
    _IN: Keyword.IN_HEADER,
    _OUT: Keyword.OUT_HEADER,
    _INTRA: Keyword.INTRA_HEADER,
}

_TRANSACTION_CLASS_TO_SECTION_NAME: Dict[Type[AbstractTransaction], Keyword] = {
    InTransaction: Keyword.IN_HEADER,
    OutTransaction: Keyword.OUT_HEADER,
    IntraTransaction: Keyword.INTRA_HEADER,
}

_TRANSACTION_CLASS_TO_TABLE: Dict[Type[AbstractTransaction], str] = {
    InTransaction: _IN,
    OutTransaction: _OUT,
    IntraTransaction: _INTRA,
}

_TABLE_TO_HEADER: Dict[str, Dict[str, str]] = {
    _IN: {
        Keyword.UNIQUE_ID.value: "Unique ID",
        Keyword.TIMESTAMP.value: "Timestamp",
        Keyword.ASSET.value: "Asset",
        Keyword.EXCHANGE.value: "Exchange",
        Keyword.HOLDER.value: "Holder",
        Keyword.TRANSACTION_TYPE.value: "Transaction Type",
        Keyword.SPOT_PRICE.value: "Spot Price",
        Keyword.CRYPTO_IN.value: "Crypto In",
        Keyword.FIAT_FEE.value: "USD Fee",
        Keyword.FIAT_IN_NO_FEE.value: "USD In No Fee",
        Keyword.FIAT_IN_WITH_FEE.value: "USD In With Fee",
        Keyword.NOTES.value: "Notes",
    },
    _OUT: {
        Keyword.UNIQUE_ID.value: "Unique ID",
        Keyword.TIMESTAMP.value: "Timestamp",
        Keyword.ASSET.value: "Asset",
        Keyword.EXCHANGE.value: "Exchange",
        Keyword.HOLDER.value: "Holder",
        Keyword.TRANSACTION_TYPE.value: "Transaction Type",
        Keyword.SPOT_PRICE.value: "Spot Price",
        Keyword.CRYPTO_OUT_NO_FEE.value: "Crypto Out No Fee",
        Keyword.CRYPTO_FEE.value: "Crypto Fee",
        Keyword.CRYPTO_OUT_WITH_FEE.value: "Crypto Out With Fee",
        Keyword.FIAT_OUT_NO_FEE.value: "USD Out No Fee",
        Keyword.FIAT_FEE.value: "USD Fee",
        Keyword.NOTES.value: "Notes",
    },
    _INTRA: {
        Keyword.UNIQUE_ID.value: "Unique ID",
        Keyword.TIMESTAMP.value: "Timestamp",
        Keyword.ASSET.value: "Asset",
        Keyword.FROM_EXCHANGE.value: "From Exchange",
        Keyword.FROM_HOLDER.value: "From Holder",
        Keyword.TO_EXCHANGE.value: "To Exchange",
        Keyword.TO_HOLDER.value: "To Holder",
        Keyword.SPOT_PRICE.value: "Spot Price",
        Keyword.CRYPTO_SENT.value: "Crypto Sent",
        Keyword.CRYPTO_RECEIVED.value: "Crypto Received",
        Keyword.NOTES.value: "Notes",
    },
}


def _transaction_sort_key(entry: AbstractTransaction) -> Tuple[str, int, datetime]:
    return (entry.asset, _TABLE_ORDER[_TRANSACTION_CLASS_TO_TABLE[entry.__class__]], entry.timestamp_value)


def generate_input_file(
    output_dir_path: str,
    output_file_prefix: str,
    output_file_name: str,
    transactions: List[AbstractTransaction],
    global_configuration: Dict[str, Any],
) -> Any:

    if not isinstance(output_dir_path, str):
        raise RP2TypeError(f"Parameter output_dir_path is not of type string: {repr(output_dir_path)}")
    if not isinstance(output_file_prefix, str):
        raise RP2TypeError(f"Parameter output_file_prefix is not of type string: {repr(output_file_prefix)}")
    if not isinstance(output_file_name, str):
        raise RP2TypeError(f"Parameter output_file_name is not of type string: {repr(output_file_name)}")
    if not isinstance(transactions, List):
        raise RP2TypeError(f"Parameter transactions is not of type List: {repr(transactions)}")

    output_file_path: Path = Path(output_dir_path) / Path(f"{output_file_prefix}{output_file_name}")
    if Path(output_file_path).exists():
        output_file_path.unlink()

    template_path: str = str(Path(os.path.dirname(__file__)).absolute() / Path("data/template.ods"))
    output_file: Any = ezodf.newdoc("ods", str(output_file_path), template=template_path)

    index: int = 0
    sheet_name: str
    sheet_indexes_to_remove: List[int] = []
    for sheet_name in output_file.sheets.names():
        if sheet_name.startswith("__"):
            # Template sheet we don't want to keep: mark it for removal
            sheet_indexes_to_remove.append(index)
        index += 1

    # Remove sheets that were marked for removal
    for index in reversed(sheet_indexes_to_remove):
        del output_file.sheets[index]

    transactions.sort(key=_transaction_sort_key)

    current_asset: Optional[str] = None
    current_sheet: Optional[Any] = None
    current_table: Optional[str] = None
    row_index: int = 0
    for transaction in transactions:
        if not isinstance(transaction, AbstractTransaction):
            raise Exception(f"Internal error: Parameter 'transaction' is not a subclass of AbstractTransaction. {transaction}")
        table_type: str = _TRANSACTION_CLASS_TO_TABLE[transaction.__class__]
        if transaction.asset != current_asset:
            if current_sheet:
                _fill_cell(current_sheet, row_index, 0, _TABLE_END, visual_style="bold")
            current_asset = transaction.asset
            current_sheet = ezodf.Table(current_asset)
            current_sheet.reset(size=(len(transactions) + _MIN_ROWS, _MAX_COLUMNS))
            output_file.sheets += current_sheet
            row_index = 0
            current_table = None
        if table_type != current_table:
            if current_table is not None:
                _fill_cell(current_sheet, row_index, 0, _TABLE_END, visual_style="bold")
                row_index += 2
            current_table = table_type
            _fill_cell(current_sheet, row_index, 0, current_table, visual_style="bold")
            _fill_header_row(current_sheet, current_table, row_index + 1, _TABLE_TO_HEADER[current_table], global_configuration)
            row_index += 2

        _fill_transaction_row(current_sheet, row_index, transaction, global_configuration)
        row_index += 1
    _fill_cell(current_sheet, row_index, 0, _TABLE_END, visual_style="bold")

    output_file.save()
    LOGGER.info("Generated output: %s", Path(output_file.docname).resolve())


def _fill_header_row(
    sheet: Any,
    current_table: str,
    row_index: int,
    header_2_name: Dict[str, str],
    global_configuration: Dict[str, Any],
) -> None:
    count: int = 0
    current_section_name: str = _TABLE_TO_SECTION_NAME[current_table].value

    for count in range(0, _MAX_COLUMNS):
        _fill_cell(sheet, row_index, count, "", visual_style="header", data_style="default")

    count = 0
    for header, name in header_2_name.items():
        column_index: int = count
        if current_section_name in global_configuration:
            if header in global_configuration[current_section_name]:
                column_index = global_configuration[current_section_name][header]
            else:
                continue
        _fill_cell(sheet, row_index, column_index, name, visual_style="header", data_style="default")
        count += 1


def _fill_transaction_row(
    sheet: Any,
    row_index: int,
    transaction: AbstractTransaction,
    global_configuration: Dict[str, Any],
) -> None:
    if not isinstance(transaction, AbstractTransaction):
        raise RP2TypeError(f"Parameter 'transaction' is not of type AbstractTransaction: {transaction}")
    count: int = 0
    current_section_name: str = _TRANSACTION_CLASS_TO_SECTION_NAME[transaction.__class__].value
    for parameter, value in transaction.constructor_parameter_dictionary.items():
        if is_internal_field(parameter):
            continue
        column_index: int = count
        if current_section_name in global_configuration:
            column_index = global_configuration[current_section_name][parameter]
        data_style: str = "default"
        if is_fiat_field(parameter):
            data_style = "fiat"
        elif is_crypto_field(parameter):
            data_style = "crypto"
        _fill_cell(sheet, row_index, column_index, value, data_style=data_style)
        count += 1


def _apply_style_to_cell(sheet: Any, row_index: int, column_index: int, style_name: str) -> None:
    sheet[row_index, column_index].style_name = style_name


def _fill_cell(
    sheet: Any,
    row_index: int,
    column_index: int,
    value: Any,
    visual_style: str = "transparent",
    data_style: str = "default",
) -> None:

    if value is None:
        return

    is_formula: bool = False
    if isinstance(value, str) and value and value[0] == "=":
        # If the value starts with '=' it is assumed to be a formula
        is_formula = True

    style_name: str = f"{visual_style}_{data_style}"
    try:
        # The ezodf API doesn't accept RP2Decimal, so we are forced to cast to float before writing to the spreadsheet
        value = float(value)
    except ValueError:
        pass
    if is_formula:
        sheet[row_index, column_index].formula = value
    else:
        sheet[row_index, column_index].set_value(value)

    _apply_style_to_cell(sheet=sheet, row_index=row_index, column_index=column_index, style_name=style_name)



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
from subprocess import run
from typing import List
import unittest

from pathlib import Path
from ods_diff import ods_diff

ROOT_PATH: Path = Path(os.path.dirname(__file__)).parent.absolute()

CONFIG_PATH: Path = ROOT_PATH / Path("config")
INPUT_PATH: Path = ROOT_PATH / Path("input")
GOLDEN_PATH: Path = INPUT_PATH / Path("golden")
OUTPUT_PATH: Path = ROOT_PATH / Path("output")

# Unit test are NYI: this is a placeholder class
class TestODSOutputDiff(unittest.TestCase):

    output_dir: Path

    @classmethod
    def setUpClass(cls) -> None:
        cls.output_dir = ROOT_PATH / Path("output") / Path(cls.__module__)
        shutil.rmtree(cls.output_dir, ignore_errors=True)

        cls._generate(OUTPUT_PATH, "test", "test_config")

    def setUp(self) -> None:  # pylint: disable=invalid-name
        self.maxDiff = None  # pylint: disable=invalid-name

    @classmethod
    def _generate(
        cls,
        output_dir: Path,
        test_name: str,
        config: str,
    ) -> None:

        arguments: List[str] = [
            "dali",
            "-s",
            "-o",
            str(output_dir),
            "-p",
            f"{test_name}_",
            str(CONFIG_PATH / Path(f"{config}.ini")),
        ]
        run(arguments, check=True)

        for method in ["fifo", "lifo"]:
            arguments = [
                "rp2_us",
                "-m",
                method,
                "-o",
                str(output_dir),
                "-p",
                f"{test_name}_",
                str(OUTPUT_PATH / Path(f"{test_name}_crypto_data.config")),
                str(OUTPUT_PATH / Path(f"{test_name}_crypto_data.ods")),
            ]
            run(arguments, check=True)

    def test_crypto_data_ods(self) -> None:
        file_name: str = "test_crypto_data.ods"
        full_output_file_name: Path = OUTPUT_PATH / file_name
        full_golden_file_name: Path = GOLDEN_PATH / file_name
        diff = ods_diff(full_golden_file_name, full_output_file_name, generate_ascii_representation=True)
        self.assertFalse(diff, msg=diff)

    def test_fifo_tax_report_us_ods(self) -> None:
        file_name: str = "test_fifo_tax_report_us.ods"
        full_output_file_name: Path = OUTPUT_PATH / file_name
        full_golden_file_name: Path = GOLDEN_PATH / file_name
        diff = ods_diff(full_golden_file_name, full_output_file_name, generate_ascii_representation=True)
        self.assertFalse(diff, msg=diff)

    def test_fifo_rp2_full_report_ods(self) -> None:
        file_name: str = "test_fifo_rp2_full_report.ods"
        full_output_file_name: Path = OUTPUT_PATH / file_name
        full_golden_file_name: Path = GOLDEN_PATH / file_name
        diff = ods_diff(full_golden_file_name, full_output_file_name, generate_ascii_representation=True)
        self.assertFalse(diff, msg=diff)

    def test_lifo_tax_report_us_ods(self) -> None:
        file_name: str = "test_lifo_tax_report_us.ods"
        full_output_file_name: Path = OUTPUT_PATH / file_name
        full_golden_file_name: Path = GOLDEN_PATH / file_name
        diff = ods_diff(full_golden_file_name, full_output_file_name, generate_ascii_representation=True)
        self.assertFalse(diff, msg=diff)

    def test_lifo_rp2_full_report_ods(self) -> None:
        file_name: str = "test_lifo_rp2_full_report.ods"
        full_output_file_name: Path = OUTPUT_PATH / file_name
        full_golden_file_name: Path = GOLDEN_PATH / file_name
        diff = ods_diff(full_golden_file_name, full_output_file_name, generate_ascii_representation=True)
        self.assertFalse(diff, msg=diff)


if __name__ == "__main__":
    unittest.main()



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from rp2.rp2_error import RP2TypeError

from dali.abstract_transaction import AbstractTransaction


class AbstractInputPlugin:
    def __init__(
        self,
        account_holder: str,
    ) -> None:
        if not isinstance(account_holder, str):
            raise RP2TypeError(f"account_holder is not a string: {account_holder}")
        self.__account_holder: str = account_holder

    def load(self) -> List[AbstractTransaction]:
        raise NotImplementedError("Abstract method: it must be implemented in the plugin class")

    @property
    def account_holder(self) -> str:
        return self.__account_holder



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from configparser import ConfigParser
from importlib import import_module
from inspect import Signature, signature
from pathlib import Path
from typing import Any, Dict, List, Set, Type, Union

from rp2.logger import LOG_FILE

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.config_generator import generate_config_file
from dali.dali_configuration import (
    DIRECTION_2_TRANSACTION_TYPE_SET,
    DEFAULT_CONFIGURATION,
    DIRECTION_SET,
    Keyword,
    is_builtin_section_name,
    is_internal_field,
    is_transaction_type_valid,
    is_unknown,
)
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.logger import LOGGER
from dali.ods_generator import generate_input_file
from dali.out_transaction import OutTransaction
from dali.transaction_resolver import DirectionTypeAndNotes, resolve_transactions

_VERSION: str = "0.3.12"


def input_loader() -> None:

    args: Namespace
    parser: ArgumentParser

    dali_configuration: Dict[str, Any] = DEFAULT_CONFIGURATION

    parser = _setup_argument_parser()
    args = parser.parse_args()

    _setup_paths(parser=parser, output_dir=args.output_dir)

    transactions: List[AbstractTransaction] = []

    try:
        ini_config: ConfigParser = ConfigParser()
        ini_config.read(args.ini_file)

        package_found: bool = False
        section_name: str
        for section_name in ini_config.sections():
            # Plugin sections can have extra trailing words to make them unique: this way there can be multiple sections
            # with the same plugin. Builtin sections cannot have extra trailing words. This split() call extracts the section
            # name and removes any trailing words.
            normalized_section_name: str = section_name.split(" ", 1)[0]
            if is_builtin_section_name(normalized_section_name):
                if section_name != normalized_section_name:
                    LOGGER.error("Builtin section '%s' cannot have extra trailing keywords: '%s'", normalized_section_name, section_name)
                    sys.exit(1)
                if section_name == Keyword.TRANSACTION_HINTS.value:
                    dali_configuration[section_name] = _validate_transaction_hints_configuration(ini_config, section_name)
                else:
                    dali_configuration[section_name] = _validate_header_configuration(ini_config, section_name)
                continue
            # Plugin section
            input_module = import_module(normalized_section_name)
            if hasattr(input_module, "InputPlugin"):
                plugin_configuration: Dict[str, Union[str, int, float, bool]] = _validate_plugin_configuration(
                    ini_config, section_name, signature(input_module.InputPlugin)
                )
                input_plugin: AbstractInputPlugin = input_module.InputPlugin(**plugin_configuration)
                LOGGER.info("Reading crypto data using plugin '%s'", section_name)
                LOGGER.debug("InputPlugin object: '%s'", input_plugin)
                if not hasattr(input_plugin, "load"):
                    LOGGER.error("Plugin '%s' has no 'load' method. Exiting...", normalized_section_name)
                    sys.exit(1)
                plugin_transactions: List[AbstractTransaction] = input_plugin.load()
                for transaction in plugin_transactions:
                    if not isinstance(transaction, AbstractTransaction):
                        LOGGER.error("Plugin '%s' returned a non-transaction object: %s. Exiting...", normalized_section_name, str(transaction))  # type: ignore
                        sys.exit(1)
                transactions.extend(plugin_transactions)
            package_found = True

        if not package_found:
            LOGGER.error("No plugin configuration found in config file. Exiting...")
            sys.exit(1)

        LOGGER.info("Resolving transactions")
        resolved_transactions: List[AbstractTransaction] = resolve_transactions(transactions, dali_configuration, args.read_spot_price_from_web)

        LOGGER.info("Generating config file in %s", args.output_dir)
        generate_config_file(args.output_dir, args.prefix, "crypto_data.config", resolved_transactions, dali_configuration)

        LOGGER.info("Generating input file in %s", args.output_dir)
        generate_input_file(args.output_dir, args.prefix, "crypto_data.ods", resolved_transactions, dali_configuration)

    except Exception:  # pylint: disable=broad-except
        LOGGER.exception("Fatal exception occurred:")

    LOGGER.info("Log file: %s", LOG_FILE)
    LOGGER.info("Generated output directory: %s", args.output_dir)
    LOGGER.info("Done")


def _setup_argument_parser() -> ArgumentParser:
    parser: ArgumentParser = ArgumentParser(
        description=(
            "Generate RP2 input and configuration files. Links:\n"
            "- documentation: https://github.com/eprbell/dali-rp2/blob/main/README.md\n"
            "- FAQ: https://github.com/eprbell/dali-rp2/blob/main/docs/user_faq.md\n"
            "- leave a star on Github: https://github.com/eprbell/dali-rp2"
        ),
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "-o",
        "--output_dir",
        action="store",
        default="output/",
        help="Write RP2 input and configuration files to OUTPUT_DIR",
        metavar="OUTPUT_DIR",
        type=str,
    )
    parser.add_argument(
        "-p",
        "--prefix",
        action="store",
        default="",
        help="Prepend RP2 input and configuration file names with PREFIX",
        metavar="PREFIX",
        type=str,
    )
    parser.add_argument(
        "-s",
        "--read-spot-price-from-web",
        action="store_true",
        help="Read spot price from Yahoo historical data (highest daily value), for transactions where it's missing",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"DaLI {_VERSION} (https://pypi.org/project/dali_rp2/)",
        help="Print DaLI version",
    )
    parser.add_argument(
        "ini_file",
        action="store",
        help="INI file",
        metavar="INI_FILE",
        type=str,
    )

    return parser


def _setup_paths(parser: ArgumentParser, output_dir: str) -> None:
    output_dir_path: Path = Path(output_dir)
    if not output_dir_path.exists():
        output_dir_path.mkdir(parents=True)
    if not output_dir_path.is_dir():
        print(f"output_dir '{output_dir}' exists but it's not a directory")
        parser.print_help()
        sys.exit(1)


# Sanity check of transaction_hints section. Ensure line format conforms to:
# <unique_id> = <direction>:<transaction_type>:<notes>
# where:
# - <unique_id> is the hash or exchange-specific id of a transaction
# - <direction> is IN/OUT,INTRA
# - <transaction_type> is a RP2 transaction type (see documentation) and
# - <notes> is an optional English sentence
def _validate_transaction_hints_configuration(ini_config: ConfigParser, section_name: str) -> Dict[str, DirectionTypeAndNotes]:
    result: Dict[str, DirectionTypeAndNotes] = {}

    if section_name != Keyword.TRANSACTION_HINTS.value:
        LOGGER.error("Invalid section name: %s", section_name)
        sys.exit(1)

    for unique_id, transaction_hint in ini_config[section_name].items():
        tokenized_transaction_hint: List[str]
        direction: str
        transaction_type: str
        notes: str
        if not unique_id or is_unknown(unique_id):
            LOGGER.error("Invalid unique id value in %s = %s", unique_id, transaction_hint)
            sys.exit(1)
        if not transaction_hint or is_unknown(transaction_hint):
            LOGGER.error("Invalid transaction_hint value in %s = %s", unique_id, transaction_hint)
            sys.exit(1)

        tokenized_transaction_hint = transaction_hint.split(":", 2)
        if len(tokenized_transaction_hint) != 3:
            LOGGER.error("Invalid transaction_hint format (expected <direction>:<transaction_type>:<notes>): %s = %s", unique_id, transaction_hint)
            sys.exit(1)

        # Check direction
        direction = tokenized_transaction_hint[0].strip().lower()
        if direction not in DIRECTION_SET:
            LOGGER.error("Direction substring in transaction_hint must be one of '%s', instead it was '%s'", str(DIRECTION_SET), direction)
            sys.exit(1)

        # Check transaction_type
        transaction_type = tokenized_transaction_hint[1].strip().lower()
        if not is_transaction_type_valid(direction, transaction_type):
            LOGGER.error(
                "Transaction type substring for '%s' direction in transaction_hint must be one of '%s', instead it was '%s'",
                direction,
                str(DIRECTION_2_TRANSACTION_TYPE_SET[direction]),
                transaction_type,
            )
            sys.exit(1)

        notes = tokenized_transaction_hint[2].strip()

        result[unique_id] = DirectionTypeAndNotes(direction, transaction_type, notes)
    return result


# Check that header section parameters map 1-1 against respective transaction constructor signature and build an parameter-to-column dictionary
def _validate_header_configuration(ini_config: ConfigParser, section_name: str) -> Dict[str, int]:
    transaction_class: Union[Type[InTransaction], Type[OutTransaction], Type[IntraTransaction]]
    result: Dict[str, int] = {}

    if section_name == Keyword.IN_HEADER.value:
        transaction_class = InTransaction
    elif section_name == Keyword.OUT_HEADER.value:
        transaction_class = OutTransaction
    elif section_name == Keyword.INTRA_HEADER.value:
        transaction_class = IntraTransaction
    else:
        LOGGER.error("Invalid section name: %s", section_name)
        sys.exit(1)

    constructor_parameter_set: Set[str] = set()
    section_parameter_set: Set[str] = set()

    for parameter in signature(transaction_class).parameters:
        if is_internal_field(parameter):
            continue
        constructor_parameter_set.add(parameter)

    column_to_header: Dict[int, str] = {}
    for header, column in ini_config[section_name].items():
        try:
            column_value: int = int(column)
        except ValueError:
            LOGGER.error("In section %s, %s has non-integer value %s", section_name, header, column)
            sys.exit(1)
        result[header] = ini_config.getint(section_name, header)
        section_parameter_set.add(header)
        if column_value in column_to_header:
            LOGGER.error("In section %s both %s and %s have the same value %s", section_name, column_to_header[column_value], header, column_value)
            sys.exit(1)
        column_to_header[column_value] = header

    # Ensure section parameters correspond 1-1 to transaction constructor parameters
    if section_parameter_set != constructor_parameter_set:
        LOGGER.error(
            "Invalid parameter set in section '%s': expected '%s', but received '%s'", section_name, str(constructor_parameter_set), str(section_parameter_set)
        )
        sys.exit(1)

    return result


# Typecheck plugin section parameters against the plugin constructor signature and build an initialization parameter dictionary
def _validate_plugin_configuration(ini_config: ConfigParser, plugin_name: str, constructor_signature: Signature) -> Dict[str, Union[str, int, float, bool]]:
    result: Dict[str, Union[str, int, float, bool]] = {}

    for parameter in constructor_signature.parameters:
        annotation: Any = constructor_signature.parameters[parameter].annotation
        if annotation is str:
            result[parameter] = ini_config[plugin_name][parameter]
        elif annotation is int:
            result[parameter] = ini_config.getint(plugin_name, parameter)
        elif annotation is float:
            result[parameter] = ini_config.getfloat(plugin_name, parameter)
        elif annotation is bool:
            result[parameter] = ini_config.getboolean(plugin_name, parameter)
        else:
            LOGGER.error("Unsupported type for parameter '%s' in plugin '%s' constructor (only str, int, float and bool are allowed)", parameter, plugin_name)
            sys.exit(1)

    return result



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# REST API: https://docs.cloud.coinbase.com/exchange/reference
# Authentication: https://docs.cloud.coinbase.com/exchange/docs/authorization-and-authentication
# Endpoint: https://api.pro.coinbase.com

import base64
import hashlib
import hmac
import json
import logging
import time
from typing import Any, Dict, List, NamedTuple, Optional, cast

import requests
from requests import PreparedRequest
from requests.auth import AuthBase
from requests.models import Response
from requests.sessions import Session
from rp2.logger import create_logger
from rp2.rp2_decimal import RP2Decimal, ZERO

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.dali_configuration import Keyword, is_fiat
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction
from dali.transaction_resolver import AssetAndUniqueId

_AMOUNT: str = "amount"
_BUY: str = "buy"
_COINBASE_TRANSACTION_ID: str = "coinbase_transaction_id"
_CREATED_AT: str = "created_at"
_CRYPTO_TRANSACTION_HASH: str = "crypto_transaction_hash"
_CURRENCY: str = "currency"
_DEPOSIT: str = "deposit"
_DETAILS: str = "details"
_FEE: str = "fee"
_ID: str = "id"
_MATCH: str = "match"
_ORDER_ID: str = "order_id"
_PRICE: str = "price"
_PRODUCT_ID: str = "product_id"
_SELL: str = "sell"
_SIDE: str = "side"
_SIZE: str = "size"
_TRADE_ID: str = "trade_id"
_TRANSFER: str = "transfer"
_TRANSFER_ID: str = "transfer_id"
_TRANSFER_TYPE: str = "transfer_type"
_TYPE: str = "type"
_USD_VOLUME: str = "usd_volume"
_WITHDRAW: str = "withdraw"


class FromToCurrencyPair(NamedTuple):
    from_currency: str
    to_currency: str


class CoinbaseProAuth(AuthBase):
    def __init__(self, api_key: str, api_secret: str, api_passphrase: str) -> None:
        self.__api_key: str = api_key
        self.__api_secret: str = api_secret
        self.__api_passphrase: str = api_passphrase

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        timestamp: str = str(time.time())
        message: str = f"{timestamp}{request.method}{request.path_url}{(request.body or '')}"  # type: ignore
        hmac_key: bytes = base64.b64decode(self.__api_secret)
        signature: hmac.HMAC = hmac.new(hmac_key, message.encode("ascii"), hashlib.sha256)
        signature_b64: str = base64.b64encode(signature.digest()).decode("utf-8")

        request.headers.update(
            {
                "Content-Type": "Application/JSON",
                "CB-ACCESS-SIGN": signature_b64,
                "CB-ACCESS-TIMESTAMP": timestamp,
                "CB-ACCESS-KEY": self.__api_key,
                "CB-ACCESS-PASSPHRASE": self.__api_passphrase,
            }
        )
        return request


class InputPlugin(AbstractInputPlugin):

    __API_URL: str = "https://api.pro.coinbase.com/"
    __TIMEOUT: int = 30

    __COINBASE_PRO: str = "Coinbase Pro"

    # Sometimes Coinbase Pro reports the same transaction twice (e.g. when swapping from a coin to another):
    # in such cases we need to ensure we add the transaction only once, hence the cache
    __fill_cache: Dict[AssetAndUniqueId, AbstractTransaction] = {}

    def __init__(
        self,
        account_holder: str,
        api_key: str,
        api_secret: str,
        api_passphrase: str,
    ) -> None:

        super().__init__(account_holder)
        self.__api_url: str = InputPlugin.__API_URL
        self.__auth = CoinbaseProAuth(api_key, api_secret, api_passphrase)
        self.__session: Session = requests.Session()
        self.__logger: logging.Logger = create_logger(f"{self.__COINBASE_PRO}/{self.account_holder}")

    def load(self) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        product_id_2_trade_id_2_fill: Dict[str, Dict[str, Any]] = {}

        for account in self.__get_accounts():
            currency: str = account[_CURRENCY]
            account_id: str = account[_ID]
            in_transaction_list: List[InTransaction] = []
            out_transaction_list: List[OutTransaction] = []
            intra_transaction_list: List[IntraTransaction] = []

            self.__logger.debug("Account: %s", json.dumps(account))
            for transaction in self.__get_transactions(account_id):
                transaction_type: str = transaction[_TYPE]
                self.__logger.debug("Transaction: %s", json.dumps(transaction))
                if transaction_type == _TRANSFER:
                    self._process_transfer(transaction, currency, intra_transaction_list)
                elif transaction_type == _MATCH:
                    product_id: str = transaction[_DETAILS][_PRODUCT_ID]
                    if product_id not in product_id_2_trade_id_2_fill:
                        trade_id_2_fill: Dict[str, Any] = {}
                        for fill in self.__get_fills(product_id):
                            trade_id_2_fill[f"{fill[_ORDER_ID]}/{fill[_TRADE_ID]}"] = fill
                        product_id_2_trade_id_2_fill[product_id] = trade_id_2_fill
                    self._process_fills(
                        transaction,
                        in_transaction_list,
                        out_transaction_list,
                        product_id_2_trade_id_2_fill[product_id][f"{transaction[_DETAILS][_ORDER_ID]}/{transaction[_DETAILS][_TRADE_ID]}"],
                    )

            if in_transaction_list:
                result.extend(in_transaction_list)
            if out_transaction_list:
                result.extend(out_transaction_list)
            if intra_transaction_list:
                result.extend(intra_transaction_list)

        return result

    @staticmethod
    def _parse_product_id(product_id: str) -> FromToCurrencyPair:
        split_product_id: List[str] = product_id.split("-")
        return FromToCurrencyPair(from_currency=split_product_id[0], to_currency=split_product_id[1])

    def _process_transfer(self, transaction: Any, currency: str, intra_transaction_list: List[IntraTransaction]) -> None:
        # Ensure the amount is positive
        amount: RP2Decimal = RP2Decimal(transaction[_AMOUNT])
        if amount < ZERO:
            amount = -amount  # type: ignore
        transaction_details: Any = transaction[_DETAILS]
        transfer_id: str = transaction_details[_TRANSFER_ID]
        transfer: Any = self.__get_transfer(transfer_id)
        transfer_details: Any = transfer[_DETAILS]
        crypto_hash: str = Keyword.UNKNOWN.value

        self.__logger.debug("Transfer: %s", json.dumps(transfer))
        if _CRYPTO_TRANSACTION_HASH not in transfer_details:
            self.__logger.debug("Transfer to/from Coinbase already captured by Coinbase plugin: ignoring.")
            return

        if _COINBASE_TRANSACTION_ID in transfer_details:
            self.__logger.debug("Transfer is a Coinbase transaction already captured by Coinbase plugin: ignoring.")
            return

        crypto_hash = transfer_details[_CRYPTO_TRANSACTION_HASH]

        if transaction_details[_TRANSFER_TYPE] == _DEPOSIT:
            intra_transaction_list.append(
                IntraTransaction(
                    self.__COINBASE_PRO,
                    crypto_hash,
                    json.dumps(transaction),
                    transaction[_CREATED_AT],
                    currency,
                    Keyword.UNKNOWN.value,
                    Keyword.UNKNOWN.value,
                    self.__COINBASE_PRO,
                    self.account_holder,
                    None,  # Spot price
                    Keyword.UNKNOWN.value,
                    str(amount),  # Received amount
                )
            )
        elif transaction_details[_TRANSFER_TYPE] == _WITHDRAW:
            intra_transaction_list.append(
                IntraTransaction(
                    self.__COINBASE_PRO,
                    crypto_hash,
                    json.dumps(transaction),
                    transaction[_CREATED_AT],
                    currency,
                    self.__COINBASE_PRO,
                    self.account_holder,
                    Keyword.UNKNOWN.value,
                    Keyword.UNKNOWN.value,
                    None,  # Spot price
                    str(amount),
                    Keyword.UNKNOWN.value,  # Received amount
                )
            )
        else:
            self.__logger.debug("Unsupported transaction type (skipping): %s", transaction_details[_TRANSFER_TYPE])

    def _process_fills(self, transaction: Any, in_transaction_list: List[InTransaction], out_transaction_list: List[OutTransaction], fill: Any) -> None:
        product_id: str = transaction[_DETAILS][_PRODUCT_ID]
        self.__logger.debug("Product id: %s", product_id)

        from_currency: str
        to_currency: str
        usd_volume: RP2Decimal
        fill_side: str = fill[_SIDE]
        unique_id: str = f"{fill[_ORDER_ID]}/{fill[_TRADE_ID]}"
        from_currency_size: RP2Decimal
        from_currency_price: RP2Decimal
        to_currency_size: RP2Decimal
        crypto_fee: RP2Decimal
        to_currency_price: RP2Decimal
        spot_price: RP2Decimal = RP2Decimal(fill[_PRICE])
        crypto_amount: RP2Decimal = RP2Decimal(fill[_SIZE])
        fiat_fee: RP2Decimal = RP2Decimal(fill[_FEE])

        from_currency, to_currency = self._parse_product_id(fill[_PRODUCT_ID])
        is_from_currency_fiat: bool = is_fiat(from_currency)
        is_to_currency_fiat: bool = is_fiat(to_currency)
        self.__logger.debug("Fill: %s", json.dumps(fill))
        if (is_from_currency_fiat and not is_to_currency_fiat and fill_side == _SELL) or (  # pylint: disable=too-many-boolean-expressions
            not is_from_currency_fiat and is_to_currency_fiat and fill_side == _BUY
        ):
            self.__append_transaction(
                cast(List[AbstractTransaction], in_transaction_list),
                cast(
                    AbstractTransaction,
                    InTransaction(
                        self.__COINBASE_PRO,
                        unique_id,
                        json.dumps(transaction),
                        fill[_CREATED_AT],
                        to_currency if is_from_currency_fiat else from_currency,
                        self.__COINBASE_PRO,
                        self.account_holder,
                        "Buy",
                        str(spot_price),
                        str(crypto_amount),
                        str(fiat_fee),
                        None,
                        None,
                        None,
                    ),
                ),
            )
        elif (is_from_currency_fiat and not is_to_currency_fiat and fill_side == _BUY) or (  # pylint: disable=too-many-boolean-expressions
            not is_from_currency_fiat and is_to_currency_fiat and fill_side == _SELL
        ):
            self.__append_transaction(
                cast(List[AbstractTransaction], out_transaction_list),
                cast(
                    AbstractTransaction,
                    OutTransaction(
                        self.__COINBASE_PRO,
                        unique_id,
                        json.dumps(transaction),
                        fill[_CREATED_AT],
                        from_currency if is_to_currency_fiat else from_currency,
                        self.__COINBASE_PRO,
                        self.account_holder,
                        "Sell",
                        str(spot_price),
                        str(crypto_amount - fiat_fee / spot_price),
                        str(fiat_fee / spot_price),
                        None,
                        None,
                        None,
                        None,
                    ),
                ),
            )
        elif not is_from_currency_fiat and not is_to_currency_fiat:
            # Convert from a crypto to another
            usd_volume = RP2Decimal(fill[_USD_VOLUME])
            crypto_fee = RP2Decimal(fill[_FEE])
            if fill_side == _SELL:
                from_currency_size = RP2Decimal(fill[_SIZE])
                from_currency_price = usd_volume / from_currency_size
                to_currency_size = from_currency_size * RP2Decimal(fill[_PRICE])
                to_currency_price = usd_volume / to_currency_size
            elif fill_side == _BUY:
                to_currency_size = RP2Decimal(fill[_SIZE])
                to_currency_price = usd_volume / to_currency_size
                from_currency_size = to_currency_size * RP2Decimal(fill[_PRICE])
                from_currency_price = usd_volume / from_currency_size
            else:
                raise Exception(f"Internal error: unsupported fill side {transaction}\n{fill}")
            self.__append_transaction(
                cast(List[AbstractTransaction], out_transaction_list),
                OutTransaction(
                    self.__COINBASE_PRO,
                    unique_id,
                    json.dumps(transaction),
                    fill[_CREATED_AT],
                    from_currency,
                    self.__COINBASE_PRO,
                    self.account_holder,
                    "Sell",
                    str(from_currency_price),
                    str(from_currency_size),
                    "0",
                    None,
                    None,
                    None,
                    f"Sell side of conversion: {from_currency_size:.8f} {from_currency} -> {to_currency_size:.8f} {to_currency}",
                ),
            )
            self.__append_transaction(
                cast(List[AbstractTransaction], in_transaction_list),
                InTransaction(
                    self.__COINBASE_PRO,
                    f"{unique_id}/buy",
                    json.dumps(transaction),
                    fill[_CREATED_AT],
                    to_currency,
                    self.__COINBASE_PRO,
                    self.account_holder,
                    "Buy",
                    str(to_currency_price),
                    str(to_currency_size),
                    "0",
                    None,
                    None,
                    f"Buy side of conversion: {from_currency_size:.8f} {from_currency} -> {to_currency_size:.8f} {to_currency}",
                ),
            )
            self.__append_transaction(
                cast(List[AbstractTransaction], out_transaction_list),
                OutTransaction(
                    self.__COINBASE_PRO,
                    f"{unique_id}/fee",
                    json.dumps(transaction),
                    fill[_CREATED_AT],
                    to_currency,
                    self.__COINBASE_PRO,
                    self.account_holder,
                    "Sell",
                    str(to_currency_price),
                    str(crypto_fee),
                    "0",
                    None,
                    None,
                    None,
                    f"Fee of conversion: {from_currency_size:.8f} {from_currency} -> {to_currency_size:.8f} {to_currency}",
                ),
            )

        else:
            self.__logger.debug("Unsupported transaction (skipping): %s", json.dumps(transaction))

    def __append_transaction(self, transaction_list: List[AbstractTransaction], transaction: AbstractTransaction) -> None:
        if AssetAndUniqueId(transaction.asset, transaction.unique_id) not in self.__fill_cache:
            transaction_list.append(transaction)
            self.__fill_cache[AssetAndUniqueId(transaction.asset, transaction.unique_id)] = transaction

    def __get_accounts(self) -> Any:
        return self.__send_request("get", "accounts")

    def __get_fills(self, product_id: str) -> Any:
        return self.__send_request_with_pagination("fills", {"product_id": f"{product_id}"})

    def __get_transfer(self, transfer_id: str) -> Any:
        return self.__send_request("get", f"transfers/{transfer_id}")

    def __get_transactions(self, account_id: str) -> Any:
        return self.__send_request_with_pagination(f"accounts/{account_id}/ledger")

    def __send_request(self, method: str, endpoint: str, params: Any = None, data: Any = None) -> Any:
        full_url: str = f"{self.__api_url}{endpoint}"
        response: Response = self.__session.request(method, full_url, params=params, data=data, auth=self.__auth, timeout=self.__TIMEOUT)
        self._validate_response(response, method, endpoint)
        return response.json()

    # Documented at: https://docs.cloud.coinbase.com/exchange/docs/pagination
    def __send_request_with_pagination(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> Any:
        if params is None:
            params = {}
        full_url: str = f"{self.__api_url}{endpoint}"
        while True:
            response: Response = self.__session.get(full_url, params=params, auth=self.__auth, timeout=self.__TIMEOUT)
            self._validate_response(response, "get", endpoint)
            json_response: Any = response.json()
            for result in json_response:
                yield result
            if not response.headers.get("cb-after"):
                break
            params["after"] = response.headers["cb-after"]

    # Documented at: https://docs.cloud.coinbase.com/exchange/docs/requests
    def _validate_response(self, response: Response, method: str, endpoint: str) -> None:
        json_response: Any = response.json()
        message: str = ""
        if 200 <= response.status_code < 300:
            return
        if "message" in json_response:
            message = json_response["message"]
            self.__logger.error("Error %d: %s%s (%s): %s", response.status_code, self.__api_url, endpoint, method.upper(), json_response["message"])

        response.raise_for_status()

        # Defensive programming: we shouldn't reach here.
        self.__logger.debug("Reached past raise_for_status() call: %s", json_response["message"])
        raise Exception(message)



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# REST API: https://developers.coinbase.com/api/v2
# Authentication: https://developers.coinbase.com/docs/wallet/api-key-authentication
# Endpoint: https://api.coinbase.com

import hashlib
import hmac
import json
import logging
import time
from typing import Any, Dict, List, Optional

import requests
from requests import PreparedRequest
from requests.auth import AuthBase
from requests.models import Response
from requests.sessions import Session
from rp2.logger import create_logger
from rp2.rp2_decimal import ZERO, RP2Decimal

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.dali_configuration import Keyword, is_fiat
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction

_AMOUNT: str = "amount"
_BUY: str = "buy"
_CODE: str = "code"
_CREATED_AT: str = "created_at"
_CURRENCY: str = "currency"
_DETAILS: str = "details"
_EMAIL: str = "email"
_EXCHANGE_DEPOSIT: str = "exchange_deposit"
_FEE: str = "fee"
_FROM: str = "from"
_HASH: str = "hash"
_ID: str = "id"
_INTEREST: str = "interest"
_NATIVE_AMOUNT: str = "native_amount"
_NETWORK: str = "network"
_OFF_BLOCKCHAIN: str = "off_blockchain"
_PRO_DEPOSIT: str = "pro_deposit"
_PRO_WITHDRAWAL: str = "pro_withdrawal"
_RESOURCE: str = "resource"
_SELL: str = "sell"
_SEND: str = "send"
_STATUS: str = "status"
_SUBTITLE: str = "subtitle"
_TO: str = "to"
_TRADE: str = "trade"
_TYPE: str = "type"
_UNIT_PRICE: str = "unit_price"
_USER: str = "user"


class CoinbaseAuth(AuthBase):

    __API_VERSION: str = "2017-11-27"

    def __init__(self, api_key: str, api_secret: str) -> None:
        self.__api_key: str = api_key
        self.__api_secret: str = api_secret

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        timestamp: str = str(int(time.time()))
        message: str = f"{timestamp}{request.method}{request.path_url}{(request.body or '')}"  # type: ignore
        secret: str = self.__api_secret
        signature: str = hmac.new(secret.encode(), message.encode(), hashlib.sha256).hexdigest()
        request.headers.update(
            {
                "CB-VERSION": self.__API_VERSION,
                "CB-ACCESS-KEY": self.__api_key,
                "CB-ACCESS-SIGN": signature,
                "CB-ACCESS-TIMESTAMP": timestamp,
            }
        )
        return request


class InputPlugin(AbstractInputPlugin):

    __API_URL: str = "https://api.coinbase.com"
    __TIMEOUT: int = 30

    __COINBASE: str = "Coinbase"
    __COINBASE_PRO: str = "Coinbase Pro"

    def __init__(
        self,
        account_holder: str,
        api_key: str,
        api_secret: str,
    ) -> None:

        super().__init__(account_holder)
        self.__api_url: str = InputPlugin.__API_URL
        self.__auth: CoinbaseAuth = CoinbaseAuth(api_key, api_secret)
        self.__session: Session = requests.Session()
        self.__logger: logging.Logger = create_logger(f"{self.__COINBASE}/{self.account_holder}")

    def load(self) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        for account in self.__get_accounts():
            currency: str = account[_CURRENCY][_CODE]
            account_id: str = account[_ID]
            in_transaction_list: List[InTransaction] = []
            out_transaction_list: List[OutTransaction] = []
            intra_transaction_list: List[IntraTransaction] = []

            self.__logger.debug("Account: %s", json.dumps(account))

            id_2_buy: Dict[str, Any] = {}
            for buy in self.__get_buys(account_id):
                id_2_buy[buy[_ID]] = buy

            id_2_sell: Dict[str, Any] = {}
            for sell in self.__get_sells(account_id):
                id_2_sell[sell[_ID]] = sell

            for transaction in self.__get_transactions(account_id):
                if is_fiat(currency):
                    self.__logger.debug("Skipping fiat transaction: %s", json.dumps(transaction))
                    continue
                self.__logger.debug("Transaction: %s", json.dumps(transaction))
                transaction_type: str = transaction[_TYPE]
                if transaction_type in {_PRO_DEPOSIT, _PRO_WITHDRAWAL, _EXCHANGE_DEPOSIT, _SEND}:
                    self._process_transfer(transaction, currency, in_transaction_list, out_transaction_list, intra_transaction_list)
                elif transaction_type in {_BUY, _SELL, _TRADE}:
                    self._process_fill(transaction, currency, in_transaction_list, out_transaction_list, id_2_buy, id_2_sell)
                elif transaction_type in {_INTEREST}:
                    self._process_interest(transaction, currency, in_transaction_list)
                else:
                    self.__logger.debug("Unsupported transaction type (skipping): %s", transaction_type)

            if in_transaction_list:
                result.extend(in_transaction_list)
            if out_transaction_list:
                result.extend(out_transaction_list)
            if intra_transaction_list:
                result.extend(intra_transaction_list)

        return result

    def _process_transfer(
        self,
        transaction: Any,
        currency: str,
        in_transaction_list: List[InTransaction],
        out_transaction_list: List[OutTransaction],
        intra_transaction_list: List[IntraTransaction],
    ) -> None:
        # Ensure the amount is positive
        amount: RP2Decimal = RP2Decimal(transaction[_AMOUNT][_AMOUNT])
        native_amount: RP2Decimal = RP2Decimal(transaction[_NATIVE_AMOUNT][_AMOUNT])
        transaction_type: str = transaction[_TYPE]

        if transaction_type == _PRO_WITHDRAWAL:
            intra_transaction_list.append(
                IntraTransaction(
                    self.__COINBASE,
                    transaction[_ID],
                    json.dumps(transaction),
                    transaction[_CREATED_AT],
                    currency,
                    self.__COINBASE_PRO,
                    self.account_holder,
                    self.__COINBASE,
                    self.account_holder,
                    None,  # Spot price
                    str(amount),
                    str(amount),
                )
            )
        elif transaction_type == _PRO_DEPOSIT:
            intra_transaction_list.append(
                IntraTransaction(
                    self.__COINBASE,
                    transaction[_ID],
                    json.dumps(transaction),
                    transaction[_CREATED_AT],
                    currency,
                    self.__COINBASE,
                    self.account_holder,
                    self.__COINBASE_PRO,
                    self.account_holder,
                    None,  # Spot price
                    str(-amount),
                    str(-amount),
                )
            )
        elif transaction_type == _EXCHANGE_DEPOSIT:
            intra_transaction_list.append(
                IntraTransaction(
                    self.__COINBASE,
                    transaction[_ID],
                    json.dumps(transaction),
                    transaction[_CREATED_AT],
                    currency,
                    self.__COINBASE,
                    self.account_holder,
                    self.__COINBASE_PRO,
                    self.account_holder,
                    None,  # Spot price
                    str(-amount),
                    str(-amount),
                )
            )
        elif transaction_type == _SEND:
            transaction_network = transaction[_NETWORK]
            crypto_hash: str = transaction_network[_HASH] if _HASH in transaction_network else Keyword.UNKNOWN.value
            if amount < ZERO:
                if (
                    transaction[_TO][_RESOURCE] == _USER
                    and transaction_network[_STATUS] == _OFF_BLOCKCHAIN
                    and _SUBTITLE in transaction[_DETAILS]
                    and _EMAIL in transaction[_TO]
                ):
                    # Outgoing gift to another Coinbase user
                    out_transaction_list.append(
                        OutTransaction(
                            self.__COINBASE,
                            transaction[_ID],
                            json.dumps(transaction),
                            transaction[_CREATED_AT],
                            currency,
                            self.__COINBASE,
                            self.account_holder,
                            "Gift",
                            str(native_amount / amount),
                            str(-amount),
                            "0",
                            str(-amount),
                            str(-native_amount),
                            "0",
                            f"To: {transaction[_TO][_EMAIL]}",
                        )
                    )
                else:
                    intra_transaction_list.append(
                        IntraTransaction(
                            self.__COINBASE,
                            crypto_hash,
                            json.dumps(transaction),
                            transaction[_CREATED_AT],
                            currency,
                            self.__COINBASE,
                            self.account_holder,
                            Keyword.UNKNOWN.value,
                            Keyword.UNKNOWN.value,
                            str(native_amount / amount),
                            str(-amount),
                            Keyword.UNKNOWN.value,
                        )
                    )
            else:
                if transaction[_FROM][_RESOURCE] == _USER and transaction_network[_STATUS] == _OFF_BLOCKCHAIN and _SUBTITLE in transaction[_DETAILS]:
                    if _EMAIL in transaction[_FROM]:
                        # Incoming gift from another Coinbase user
                        in_transaction_list.append(
                            InTransaction(
                                self.__COINBASE,
                                transaction[_ID],
                                json.dumps(transaction),
                                transaction[_CREATED_AT],
                                currency,
                                self.__COINBASE,
                                self.account_holder,
                                "Income",
                                str(native_amount / amount),
                                transaction[_AMOUNT][_AMOUNT],
                                "0",
                                str(native_amount),
                                "0",
                                f"From: {transaction[_FROM][_EMAIL]}",
                            )
                        )
                    elif transaction[_DETAILS][_SUBTITLE].startswith("From Coinbase"):
                        # Coinbase Earn transactions
                        in_transaction_list.append(
                            InTransaction(
                                self.__COINBASE,
                                transaction[_ID],
                                json.dumps(transaction),
                                transaction[_CREATED_AT],
                                currency,
                                self.__COINBASE,
                                self.account_holder,
                                "Income",
                                str(native_amount / amount),
                                transaction[_AMOUNT][_AMOUNT],
                                "0",
                                str(native_amount),
                                str(native_amount),
                                "Coinbase EARN",
                            )
                        )
                else:
                    intra_transaction_list.append(
                        IntraTransaction(
                            self.__COINBASE,
                            crypto_hash,
                            json.dumps(transaction),
                            transaction[_CREATED_AT],
                            currency,
                            Keyword.UNKNOWN.value,
                            Keyword.UNKNOWN.value,
                            self.__COINBASE,
                            self.account_holder,
                            str(native_amount / amount),
                            Keyword.UNKNOWN.value,
                            str(amount),
                        )
                    )

    def _process_fill(
        self,
        transaction: Any,
        currency: str,
        in_transaction_list: List[InTransaction],
        out_transaction_list: List[OutTransaction],
        id_2_buy: Dict[str, Any],
        id_2_sell: Dict[str, Any],
    ) -> None:
        transaction_type: str = transaction[_TYPE]
        native_amount: RP2Decimal = RP2Decimal(transaction[_NATIVE_AMOUNT][_AMOUNT])
        crypto_amount: RP2Decimal = RP2Decimal(transaction[_AMOUNT][_AMOUNT])
        fiat_fee: RP2Decimal = ZERO
        spot_price: RP2Decimal

        if transaction[_NATIVE_AMOUNT][_CURRENCY] != "USD":
            # This is probably a coin swap: TBD (for now just return)
            self.__logger.warning("Swap transaction encountered (swaps not supported yet by Coinbase plugin): %s", json.dumps(transaction))
            return

        if transaction_type in {_BUY, _TRADE}:
            spot_price = (native_amount - fiat_fee) / crypto_amount
            if transaction_type == _BUY:
                buy: Dict[str, Any] = id_2_buy[transaction[transaction_type][_ID]]
                fiat_fee = RP2Decimal(buy[_FEE][_AMOUNT])
                self.__logger.debug("Buy: %s", json.dumps(buy))
                spot_price = RP2Decimal(buy[_UNIT_PRICE][_AMOUNT])
            in_transaction_list.append(
                InTransaction(
                    self.__COINBASE,
                    transaction[_ID],
                    json.dumps(transaction),
                    transaction[_CREATED_AT],
                    currency,
                    self.__COINBASE,
                    self.account_holder,
                    "Buy",
                    str(spot_price),
                    str(crypto_amount),
                    str(fiat_fee),
                    str(native_amount - fiat_fee),
                    str(native_amount),
                    None,  # Add notes
                )
            )
        elif transaction_type == _SELL:
            sell = id_2_sell[transaction[transaction_type][_ID]]
            fiat_fee = RP2Decimal(sell[_FEE][_AMOUNT])
            spot_price = RP2Decimal(sell[_UNIT_PRICE][_AMOUNT])
            self.__logger.debug("Sell: %s", json.dumps(sell))
            out_transaction_list.append(
                OutTransaction(
                    self.__COINBASE,
                    transaction[_ID],
                    json.dumps(transaction),
                    transaction[_CREATED_AT],
                    currency,
                    self.__COINBASE,
                    self.account_holder,
                    "Sell",
                    str(spot_price),
                    str(-crypto_amount - fiat_fee / spot_price),
                    str(fiat_fee / spot_price),
                    str(-crypto_amount),
                    str(-native_amount),
                    str(fiat_fee),
                    None,  # Add notes
                )
            )
        else:
            self.__logger.debug("Unsupported transaction type (skipping): %s", json.dumps(transaction_type))

    def _process_interest(self, transaction: Any, currency: str, in_transaction_list: List[InTransaction]) -> None:

        in_transaction_list.append(
            InTransaction(
                self.__COINBASE,
                transaction[_ID],
                json.dumps(transaction),
                transaction[_CREATED_AT],
                currency,
                self.__COINBASE,
                self.account_holder,
                "Interest",
                Keyword.UNKNOWN.value,
                transaction[_AMOUNT][_AMOUNT],
                "0",  # Fee TBD
                None,
                None,
                None,  # Add notes
            )
        )

    def __get_accounts(self) -> Any:
        return self.__send_request_with_pagination("/v2/accounts")

    def __get_transactions(self, account_id: str) -> Any:
        return self.__send_request_with_pagination(f"/v2/accounts/{account_id}/transactions")

    def __get_buys(self, account_id: str) -> Any:
        return self.__send_request_with_pagination(f"/v2/accounts/{account_id}/buys")

    def __get_sells(self, account_id: str) -> Any:
        return self.__send_request_with_pagination(f"/v2/accounts/{account_id}/sells")

    def __send_request(self, method: str, endpoint: str, params: Any = None, data: Any = None) -> Any:  # pylint: disable=unused-private-member
        full_url: str = f"{self.__api_url}{endpoint}"
        response: Response = self.__session.request(method, full_url, params=params, data=data, auth=self.__auth, timeout=self.__TIMEOUT)
        self._validate_response(response, method, endpoint)
        return response.json()

    # Documented at: https://docs.cloud.coinbase.com/exchange/docs/pagination
    def __send_request_with_pagination(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> Any:
        if params is None:
            params = {}
        current_url: str = f"{self.__api_url}{endpoint}"
        while True:
            response: Response = self.__session.get(current_url, params=params, auth=self.__auth, timeout=self.__TIMEOUT)
            self._validate_response(response, "get", endpoint)
            json_response: Any = response.json()
            for result in json_response["data"]:
                yield result
            if not "pagination" in json_response or not "next_uri" in json_response["pagination"] or not json_response["pagination"]["next_uri"]:
                break
            current_url = f"{self.__api_url}{json_response['pagination']['next_uri']}"

    # Documented at: https://docs.cloud.coinbase.com/exchange/docs/requests
    def _validate_response(self, response: Response, method: str, endpoint: str) -> None:
        json_response: Any = response.json()
        message: str = ""
        if 200 <= response.status_code < 300:
            return
        if "message" in json_response:
            message = json_response["message"]
            self.__logger.error("Error %d: %s%s (%s): %s", response.status_code, self.__api_url, endpoint, method.upper(), json_response["message"])

        response.raise_for_status()

        # Defensive programming: we shouldn't reach here.
        self.__logger.debug("Reached past raise_for_status() call: %s", json_response["message"])
        raise Exception(message)



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Dict, Optional, Set

from rp2.rp2_error import RP2ValueError


# Configuration file keywords
class Keyword(Enum):
    AIRDROP: str = "airdrop"
    ASSET: str = "asset"
    BUY: str = "buy"
    CRYPTO_FEE: str = "crypto_fee"
    CRYPTO_IN: str = "crypto_in"
    CRYPTO_OUT_NO_FEE: str = "crypto_out_no_fee"
    CRYPTO_OUT_WITH_FEE: str = "crypto_out_with_fee"
    CRYPTO_RECEIVED: str = "crypto_received"
    CRYPTO_SENT: str = "crypto_sent"
    DONATE: str = "donate"
    EXCHANGE: str = "exchange"
    FIAT_FEE: str = "fiat_fee"
    FIAT_IN_NO_FEE: str = "fiat_in_no_fee"
    FIAT_IN_WITH_FEE: str = "fiat_in_with_fee"
    FIAT_OUT_NO_FEE: str = "fiat_out_no_fee"
    FROM_EXCHANGE: str = "from_exchange"
    FROM_HOLDER: str = "from_holder"
    GIFT: str = "gift"
    HARDFORK: str = "hardfork"
    HOLDER: str = "holder"
    IN: str = "in"
    INCOME: str = "income"
    INTEREST: str = "interest"
    INTRA: str = "intra"
    IN_HEADER: str = "in_header"
    INTRA_HEADER: str = "intra_header"
    IS_SPOT_PRICE_FROM_WEB: str = "is_spot_price_from_web"
    MINING: str = "mining"
    MOVE: str = "move"
    NOTES: str = "notes"
    OUT: str = "out"
    OUT_HEADER: str = "out_header"
    PLUGIN: str = "plugin"
    RAW_DATA: str = "raw_data"
    SELL: str = "sell"
    SPOT_PRICE: str = "spot_price"
    STAKING: str = "staking"
    TIMESTAMP: str = "timestamp"
    TO_EXCHANGE: str = "to_exchange"
    TO_HOLDER: str = "to_holder"
    TRANSACTION_HINTS: str = "transaction_hints"
    TRANSACTION_TYPE: str = "transaction_type"
    UNIQUE_ID: str = "unique_id"
    UNKNOWN = "__unknown"
    WAGES: str = "wages"

    @classmethod
    def has_value(cls, value: str) -> bool:
        return value in _keyword_values

    @classmethod
    def type_check_from_string(cls, keyword: str) -> "Keyword":
        if not Keyword.has_value(keyword.lower()):
            raise RP2ValueError(f"Invalid keyword: {keyword}")
        return Keyword[keyword.upper()]


_keyword_values: Set[str] = {item.value for item in Keyword}

_FIAT_SET: Set[str] = {"USD", "USDC"}

_FIAT_FIELD_SET: Set[str] = {
    Keyword.FIAT_FEE.value,
    Keyword.FIAT_IN_NO_FEE.value,
    Keyword.FIAT_IN_WITH_FEE.value,
    Keyword.FIAT_OUT_NO_FEE.value,
    Keyword.SPOT_PRICE.value,
}

_CRYPTO_FIELD_SET: Set[str] = {
    Keyword.CRYPTO_FEE.value,
    Keyword.CRYPTO_IN.value,
    Keyword.CRYPTO_OUT_NO_FEE.value,
    Keyword.CRYPTO_OUT_WITH_FEE.value,
    Keyword.CRYPTO_RECEIVED.value,
    Keyword.CRYPTO_SENT.value,
}

_INTERNAL_FIELD_SET: Set[str] = {
    Keyword.IS_SPOT_PRICE_FROM_WEB.value,
    Keyword.PLUGIN.value,
    Keyword.RAW_DATA.value,
}

DIRECTION_SET: Set[str] = {
    Keyword.IN.value,
    Keyword.OUT.value,
    Keyword.INTRA.value,
}

DIRECTION_2_TRANSACTION_TYPE_SET: Dict[str, Set[str]] = {
    Keyword.IN.value: {
        Keyword.AIRDROP.value,
        Keyword.BUY.value,
        Keyword.DONATE.value,
        Keyword.GIFT.value,
        Keyword.HARDFORK.value,
        Keyword.INCOME.value,
        Keyword.INTEREST.value,
        Keyword.MINING.value,
        Keyword.STAKING.value,
        Keyword.WAGES.value,
    },
    Keyword.OUT.value: {
        Keyword.DONATE.value,
        Keyword.GIFT.value,
        Keyword.SELL.value,
    },
    Keyword.INTRA.value: {
        Keyword.MOVE.value,
    },
}

BUILTIN_CONFIGURATION_SECTIONS: Set[str] = {Keyword.TRANSACTION_HINTS.value, Keyword.IN_HEADER.value, Keyword.OUT_HEADER.value, Keyword.INTRA_HEADER.value}

DEFAULT_CONFIGURATION: Dict[str, Dict[str, int]] = {
    Keyword.IN_HEADER.value: {
        Keyword.TIMESTAMP.value: 0,
        Keyword.ASSET.value: 1,
        Keyword.EXCHANGE.value: 2,
        Keyword.HOLDER.value: 3,
        Keyword.TRANSACTION_TYPE.value: 4,
        Keyword.SPOT_PRICE.value: 6,
        Keyword.CRYPTO_IN.value: 7,
        Keyword.FIAT_FEE.value: 8,
        Keyword.FIAT_IN_NO_FEE.value: 9,
        Keyword.FIAT_IN_WITH_FEE.value: 10,
        Keyword.UNIQUE_ID.value: 12,
        Keyword.NOTES.value: 13,
    },
    Keyword.OUT_HEADER.value: {
        Keyword.TIMESTAMP.value: 0,
        Keyword.ASSET.value: 1,
        Keyword.EXCHANGE.value: 2,
        Keyword.HOLDER.value: 3,
        Keyword.TRANSACTION_TYPE.value: 4,
        Keyword.SPOT_PRICE.value: 6,
        Keyword.CRYPTO_OUT_NO_FEE.value: 7,
        Keyword.CRYPTO_FEE.value: 8,
        Keyword.CRYPTO_OUT_WITH_FEE.value: 9,
        Keyword.FIAT_OUT_NO_FEE.value: 10,
        Keyword.FIAT_FEE.value: 11,
        Keyword.UNIQUE_ID.value: 12,
        Keyword.NOTES.value: 13,
    },
    Keyword.INTRA_HEADER.value: {
        Keyword.TIMESTAMP.value: 0,
        Keyword.ASSET.value: 1,
        Keyword.FROM_EXCHANGE.value: 2,
        Keyword.FROM_HOLDER.value: 3,
        Keyword.TO_EXCHANGE.value: 4,
        Keyword.TO_HOLDER.value: 5,
        Keyword.SPOT_PRICE.value: 6,
        Keyword.CRYPTO_SENT.value: 7,
        Keyword.CRYPTO_RECEIVED.value: 8,
        Keyword.UNIQUE_ID.value: 12,
        Keyword.NOTES.value: 13,
    },
}


def is_builtin_section_name(section_name: str) -> bool:
    return section_name in BUILTIN_CONFIGURATION_SECTIONS


def is_fiat_field(field: str) -> bool:
    return field in _FIAT_FIELD_SET


def is_crypto_field(field: str) -> bool:
    return field in _CRYPTO_FIELD_SET


def is_internal_field(field: str) -> bool:
    return field in _INTERNAL_FIELD_SET


def is_fiat(currency: str) -> bool:
    return currency in _FIAT_SET


def is_unknown(value: str) -> bool:
    return value == Keyword.UNKNOWN.value


def is_unknown_or_none(value: Optional[str]) -> bool:
    return value in {Keyword.UNKNOWN.value, None}


def is_transaction_type_valid(direction: str, transaction_type: str) -> bool:
    return transaction_type.lower() in DIRECTION_2_TRANSACTION_TYPE_SET[direction]



# Copyright 2022 eprbell
# Licensed under the Apache License, Version 2.0 (the "License");
#
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime
from typing import Any, Dict, List, NamedTuple, Optional

import yfinance as yf  # type: ignore
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2ValueError, RP2TypeError

from dali.abstract_transaction import AbstractTransaction
from dali.dali_configuration import Keyword, is_unknown, is_unknown_or_none
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.logger import LOGGER
from dali.out_transaction import OutTransaction

__RESOLVER: str = "DaLI Resolver"


class AssetAndUniqueId(NamedTuple):
    asset: str
    unique_id: str


class DirectionTypeAndNotes(NamedTuple):
    direction: str
    transaction_type: str
    notes: str


def _is_number(value: str) -> bool:
    try:
        float(value)
        return True
    except ValueError:
        return False


def _resolve_fields(
    name1: str,
    name2: str,
    value1: str,
    value2: str,
    transaction1: AbstractTransaction,
    transaction2: AbstractTransaction,
    disallow_two_unknown: bool = True,
    if_conflict_override_second_parameter: bool = False,
) -> str:
    if disallow_two_unknown and is_unknown(value1) and is_unknown(value2):
        raise Exception(f"Internal error: {name1} and {name2} are both unknown: {transaction1}\n{transaction2}")
    if value1 and value2 and not is_unknown(value1) and not is_unknown(value2) and value1 != value2:
        if if_conflict_override_second_parameter:
            return value1
        # If values are numeric we need to compare them as numbers, not as strings
        if _is_number(value1) and _is_number(value2):
            if RP2Decimal(value1) != RP2Decimal(value2):
                raise Exception(f"Internal error: {name1} and {name2} are numeric but have different values: {transaction1}\n{transaction2}")
        else:
            raise Exception(f"Internal error: {name1} and {name2} have different values: {transaction1}\n{transaction2}")
    if is_unknown(value1):
        return value2
    if is_unknown(value2):
        return value1
    if not value1:
        return value2
    if not value2:
        return value1
    return value1


def _resolve_optional_fields(
    name1: str,
    name2: str,
    value1: Optional[str],
    value2: Optional[str],
    transaction1: AbstractTransaction,
    transaction2: AbstractTransaction,
    disallow_two_unknown: bool = True,
    if_conflict_override_second_parameter: bool = False,
) -> str:
    if is_unknown_or_none(value1) and is_unknown_or_none(value2):
        value1 = Keyword.UNKNOWN.value
        value2 = Keyword.UNKNOWN.value
    return _resolve_fields(
        name1,
        name2,
        value1 if value1 else "",
        value2 if value2 else "",
        transaction1,
        transaction2,
        disallow_two_unknown,
        if_conflict_override_second_parameter,
    )


def _update_spot_price_from_web(transaction: AbstractTransaction) -> AbstractTransaction:
    init_parameters: Dict[str, Any] = transaction.constructor_parameter_dictionary
    if is_unknown(transaction.spot_price):  # type: ignore
        ticker: yf.Ticker = yf.Ticker(f"{transaction.asset}-USD")
        timestamp: str = transaction.timestamp_value.strftime("%Y-%m-%d")
        spot_price: str = str(ticker.history(start=timestamp, end=timestamp)["High"][0])
        notes: str = ""
        if transaction.notes:
            notes = f"{notes}; "
        notes = f"spot_price read from Yahoo Finance; {transaction.notes if transaction.notes else ''}"
        init_parameters[Keyword.SPOT_PRICE.value] = spot_price
        init_parameters[Keyword.NOTES.value] = notes
        init_parameters[Keyword.IS_SPOT_PRICE_FROM_WEB.value] = True
        if isinstance(transaction, InTransaction):
            return InTransaction(**init_parameters)
        if isinstance(transaction, OutTransaction):
            return OutTransaction(**init_parameters)
        if isinstance(transaction, IntraTransaction):
            return IntraTransaction(**init_parameters)
        raise Exception(f"Invalid transaction: {transaction}")

    return transaction


# This resolves incomplete transactions. An INTRA transaction from an exchange (e.g. Coinbase) to another
# (e.g. BlockFi) would be represented by two transactions (one created by the Coinbase plugin and the other by
# the BlockFi plugin). These two transactions may have incomplete information (e.g. Coinbase doesn't know the
# receiver address is BlockFi's), so this function matches the two incomplete transactions using their hash
# and/or account-specific id and merges them into one complete transaction.
def resolve_transactions(
    transactions: List[AbstractTransaction],
    global_configuration: Dict[str, Any],
    read_spot_price_from_web: bool,
) -> List[AbstractTransaction]:
    if not isinstance(transactions, List):
        raise Exception(f"Internal error: parameter 'transactions' is not of type List. {transactions}")

    resolved_transactions: List[AbstractTransaction] = []
    unique_id_2_transactions: Dict[AssetAndUniqueId, List[AbstractTransaction]] = {}
    transaction: AbstractTransaction

    for transaction in transactions:
        if not isinstance(transaction, AbstractTransaction):
            raise Exception(f"Internal error: Parameter 'transaction' is not a subclass of AbstractTransaction. {transaction}")

        if is_unknown(transaction.unique_id):
            # Cannot resolve further if unique_id is not known
            if read_spot_price_from_web:
                transaction = _update_spot_price_from_web(transaction)
            LOGGER.debug("Unresolvable transaction (no %s): %s", Keyword.UNIQUE_ID.value, str(transaction))
            resolved_transactions.append(transaction)
        else:
            transaction_list: List[AbstractTransaction] = unique_id_2_transactions.setdefault(AssetAndUniqueId(transaction.asset, transaction.unique_id), [])
            transaction_list.append(transaction)
            unique_id_2_transactions[AssetAndUniqueId(transaction.asset, transaction.unique_id)] = transaction_list

    for transaction_list in unique_id_2_transactions.values():
        if len(transaction_list) > 2:
            raise Exception(f"Internal error: Attempting to resolve more than two transactions with same {Keyword.UNIQUE_ID.value}: {transaction_list}")
        if len(transaction_list) == 1:
            transaction = _apply_transaction_hint(transaction_list[0], global_configuration)
            if read_spot_price_from_web:
                transaction = _update_spot_price_from_web(transaction)
            LOGGER.debug("Self-contained transaction: %s", str(transaction))
            resolved_transactions.append(transaction)
            continue
        if len(transaction_list) == 0:
            raise Exception(f"Internal error: Attempting to resolve zero transactions: {transaction_list}")

        transaction1: AbstractTransaction = transaction_list[0]
        transaction2: AbstractTransaction = transaction_list[1]

        if transaction1.unique_id != transaction2.unique_id:
            raise Exception(f"Internal error: transaction1.{Keyword.UNIQUE_ID.value} != transaction2.{Keyword.UNIQUE_ID.value}: {transaction1}\n{transaction2}")
        if transaction1.asset != transaction2.asset:
            raise Exception(f"Internal error: transaction1.{Keyword.ASSET.value} != transaction2.{Keyword.ASSET.value}: {transaction1}\n{transaction2}")

        if isinstance(transaction1, InTransaction) and isinstance(transaction2, OutTransaction):
            transaction = _resolve_in_out_transaction(transaction1, transaction2, None)
        elif isinstance(transaction1, OutTransaction) and isinstance(transaction2, InTransaction):
            transaction = _resolve_out_in_transaction(transaction1, transaction2, None)
        elif isinstance(transaction1, IntraTransaction) and isinstance(transaction2, IntraTransaction):
            transaction = _resolve_intra_intra_transaction(transaction1, transaction2, None)
        else:
            raise Exception(
                f"Internal error: attempting to resolve two transactions that aren't Intra/Intra, In/Out or Out/In:\n{transaction1}\n{transaction2}"
            )

        if read_spot_price_from_web:
            transaction = _update_spot_price_from_web(transaction)

        LOGGER.debug("Resolved transaction: %s", str(transaction))
        resolved_transactions.append(transaction)

    return resolved_transactions


def _apply_transaction_hint(
    transaction: AbstractTransaction,
    global_configuration: Dict[str, Any],
) -> AbstractTransaction:
    result: AbstractTransaction
    if transaction.unique_id not in global_configuration[Keyword.TRANSACTION_HINTS.value]:
        return transaction
    direction: str
    transaction_type: str
    notes: str
    (direction, transaction_type, notes) = global_configuration[Keyword.TRANSACTION_HINTS.value][transaction.unique_id]
    transaction_type = transaction_type.capitalize()
    notes = f"{notes}; {transaction.notes if transaction.notes else ''}"
    if direction == Keyword.IN.value:
        if isinstance(transaction, InTransaction):
            result = InTransaction(
                plugin=transaction.plugin,
                unique_id=transaction.unique_id,
                raw_data=f"{Keyword.IN.value}->{Keyword.IN.value}: {transaction.raw_data}",
                timestamp=transaction.timestamp,
                asset=transaction.asset,
                exchange=transaction.exchange,
                holder=transaction.holder,
                transaction_type=transaction_type,
                spot_price=transaction.spot_price if transaction.spot_price else Keyword.UNKNOWN.value,
                crypto_in=transaction.crypto_in,
                fiat_fee=transaction.crypto_in,
                notes=notes,
            )
        elif isinstance(transaction, OutTransaction):
            raise RP2TypeError(f"Cannot change OutTransaction to InTransaction: {transaction}")
        elif isinstance(transaction, IntraTransaction):
            if not is_unknown(transaction.from_holder) or not is_unknown(transaction.from_exchange):
                raise RP2ValueError(
                    (
                        f"Invalid conversion {Keyword.INTRA.value}->{Keyword.IN.value}: "
                        f"{Keyword.FROM_HOLDER.value}/{Keyword.FROM_EXCHANGE.value} must be unknown: {transaction}"
                    )
                )
            result = InTransaction(
                plugin=transaction.plugin,
                unique_id=transaction.unique_id,
                raw_data=f"{Keyword.INTRA.value}->{Keyword.IN.value}: {transaction.raw_data}",
                timestamp=transaction.timestamp,
                asset=transaction.asset,
                exchange=transaction.to_exchange,
                holder=transaction.to_holder,
                transaction_type=transaction_type,
                spot_price=transaction.spot_price if transaction.spot_price else Keyword.UNKNOWN.value,
                crypto_in=transaction.crypto_received,
                fiat_fee="0",
                notes=notes,
            )
    elif direction == Keyword.OUT.value:
        if isinstance(transaction, InTransaction):
            raise RP2TypeError(f"Cannot change InTransaction to OutTransaction: {transaction}")
        if isinstance(transaction, OutTransaction):
            result = OutTransaction(
                plugin=transaction.plugin,
                unique_id=transaction.unique_id,
                raw_data=f"{Keyword.OUT.value}->{Keyword.OUT.value}: {transaction.raw_data}",
                timestamp=transaction.timestamp,
                asset=transaction.asset,
                exchange=transaction.exchange,
                holder=transaction.holder,
                transaction_type=transaction_type,
                spot_price=transaction.spot_price if transaction.spot_price else Keyword.UNKNOWN.value,
                crypto_out_no_fee=transaction.crypto_out_no_fee,
                crypto_fee=transaction.crypto_fee,
                notes=notes,
            )
        elif isinstance(transaction, IntraTransaction):
            if not is_unknown(transaction.to_holder) or not is_unknown(transaction.to_exchange):
                raise RP2ValueError(
                    (
                        f"Invalid converstion {Keyword.INTRA.value}->{Keyword.OUT.value}: "
                        f"{Keyword.TO_HOLDER.value}/{Keyword.TO_EXCHANGE.value} must be unknown: {transaction}"
                    )
                )
            result = OutTransaction(
                plugin=transaction.plugin,
                unique_id=transaction.unique_id,
                raw_data=f"{Keyword.INTRA.value}->{Keyword.OUT.value}: {transaction.raw_data}",
                timestamp=transaction.timestamp,
                asset=transaction.asset,
                exchange=transaction.from_exchange,
                holder=transaction.from_holder,
                transaction_type=transaction_type,
                spot_price=transaction.spot_price if transaction.spot_price else Keyword.UNKNOWN.value,
                crypto_out_no_fee=transaction.crypto_sent,
                crypto_fee="0",
                notes=notes,
            )
    elif direction == Keyword.INTRA.value:
        if isinstance(transaction, InTransaction):
            result = IntraTransaction(
                plugin=transaction.plugin,
                unique_id=transaction.unique_id,
                raw_data=f"{Keyword.IN.value}->{Keyword.INTRA.value}: {transaction.raw_data}",
                timestamp=transaction.timestamp,
                asset=transaction.asset,
                from_exchange=Keyword.UNKNOWN.value,
                from_holder=Keyword.UNKNOWN.value,
                to_exchange=transaction.exchange,
                to_holder=transaction.holder,
                spot_price=transaction.spot_price,
                crypto_sent=Keyword.UNKNOWN.value,
                crypto_received=transaction.crypto_in,
                notes=notes,
            )
        elif isinstance(transaction, OutTransaction):
            result = IntraTransaction(
                plugin=transaction.plugin,
                unique_id=transaction.unique_id,
                raw_data=f"{Keyword.OUT.value}->{Keyword.INTRA.value}: {transaction.raw_data}",
                timestamp=transaction.timestamp,
                asset=transaction.asset,
                from_exchange=transaction.exchange,
                from_holder=transaction.holder,
                to_exchange=Keyword.UNKNOWN.value,
                to_holder=Keyword.UNKNOWN.value,
                spot_price=transaction.spot_price,
                crypto_sent=str(RP2Decimal(transaction.crypto_out_no_fee) + RP2Decimal(transaction.crypto_fee)),
                crypto_received=Keyword.UNKNOWN.value,
                notes=notes,
            )
        elif isinstance(transaction, IntraTransaction):
            result = IntraTransaction(
                plugin=transaction.plugin,
                unique_id=transaction.unique_id,
                raw_data=f"{Keyword.INTRA.value}->{Keyword.INTRA.value}: {transaction.raw_data}",
                timestamp=transaction.timestamp,
                asset=transaction.asset,
                from_exchange=transaction.from_exchange,
                from_holder=transaction.from_holder,
                to_exchange=transaction.to_exchange,
                to_holder=transaction.to_holder,
                spot_price=transaction.spot_price,
                crypto_sent=transaction.crypto_sent,
                crypto_received=transaction.crypto_received,
                notes=notes,
            )
    else:
        raise RP2ValueError(f"Invalid direction {direction}")

    return result


def _resolve_intra_intra_transaction(
    transaction1: IntraTransaction,
    transaction2: IntraTransaction,
    notes: Optional[str],
) -> IntraTransaction:

    # Pick the max of the two timestamps as the timestamp of the new resolved transaction
    timestamp: datetime = max(transaction1.timestamp_value, transaction2.timestamp_value)
    from_exchange: str = _resolve_fields(
        f"transaction1.{Keyword.FROM_EXCHANGE.value}",
        f"transaction2.{Keyword.FROM_EXCHANGE.value}",
        transaction1.from_exchange,
        transaction2.from_exchange,
        transaction1,
        transaction2,
    )
    from_holder: str = _resolve_fields(
        f"transaction1.{Keyword.FROM_HOLDER.value}",
        f"transaction2.{Keyword.FROM_HOLDER.value}",
        transaction1.from_holder,
        transaction2.from_holder,
        transaction1,
        transaction2,
    )
    to_exchange: str = _resolve_fields(
        f"transaction1.{Keyword.TO_EXCHANGE.value}",
        f"transaction2.{Keyword.TO_EXCHANGE.value}",
        transaction1.to_exchange,
        transaction2.to_exchange,
        transaction1,
        transaction2,
    )
    to_holder: str = _resolve_fields(
        f"transaction1.{Keyword.TO_HOLDER.value}",
        f"transaction2.{Keyword.TO_HOLDER.value}",
        transaction1.to_holder,
        transaction2.to_holder,
        transaction1,
        transaction2,
    )
    spot_price: str
    # If one of the transaction has spot_price from web, and the other doesn't take the non-web one
    if transaction1.is_spot_price_from_web:
        spot_price = _resolve_optional_fields(
            f"transaction2.{Keyword.SPOT_PRICE.value}",
            f"transaction1.{Keyword.SPOT_PRICE.value}",
            transaction2.spot_price,
            transaction1.spot_price,
            transaction2,
            transaction1,
            disallow_two_unknown=False,
            if_conflict_override_second_parameter=True,
        )
    elif transaction2.is_spot_price_from_web:
        spot_price = _resolve_optional_fields(
            f"transaction1.{Keyword.SPOT_PRICE.value}",
            f"transaction2.{Keyword.SPOT_PRICE.value}",
            transaction1.spot_price,
            transaction2.spot_price,
            transaction1,
            transaction2,
            disallow_two_unknown=False,
            if_conflict_override_second_parameter=True,
        )
    else:
        spot_price = _resolve_optional_fields(
            f"transaction1.{Keyword.SPOT_PRICE.value}",
            f"transaction2.{Keyword.SPOT_PRICE.value}",
            transaction1.spot_price,
            transaction2.spot_price,
            transaction1,
            transaction2,
            disallow_two_unknown=False,
            if_conflict_override_second_parameter=False,
        )
    crypto_sent: str = _resolve_fields(
        f"transaction1.{Keyword.CRYPTO_SENT.value}",
        f"transaction2.{Keyword.CRYPTO_SENT.value}",
        transaction1.crypto_sent,
        transaction2.crypto_sent,
        transaction1,
        transaction2,
    )
    crypto_received: str = _resolve_fields(
        f"transaction1.{Keyword.CRYPTO_RECEIVED.value}",
        f"transaction2.{Keyword.CRYPTO_RECEIVED.value}",
        transaction1.crypto_received,
        transaction2.crypto_received,
        transaction1,
        transaction2,
    )

    notes = f"{notes}; " if notes else ""
    notes += f"{transaction1.notes}; " if transaction1.notes else ""
    notes += f"{transaction2.notes}; " if transaction2.notes else ""

    return IntraTransaction(
        plugin=__RESOLVER,
        unique_id=transaction1.unique_id,
        raw_data=f"{transaction1.raw_data}\n{transaction2.raw_data}",
        asset=transaction1.asset,
        timestamp=str(timestamp),
        from_exchange=from_exchange,
        from_holder=from_holder,
        to_exchange=to_exchange,
        to_holder=to_holder,
        spot_price=spot_price,
        crypto_sent=crypto_sent,
        crypto_received=crypto_received,
        notes=notes,
    )


def _resolve_out_in_transaction(
    out_transaction: OutTransaction,
    in_transaction: InTransaction,
    notes: Optional[str],
) -> IntraTransaction:
    return _resolve_in_out_transaction(in_transaction, out_transaction, notes)


def _resolve_in_out_transaction(
    in_transaction: InTransaction,
    out_transaction: OutTransaction,
    notes: Optional[str],
) -> IntraTransaction:

    timestamp: datetime = in_transaction.timestamp_value
    from_exchange: str = out_transaction.exchange
    from_holder: str = out_transaction.holder
    to_exchange: str = in_transaction.exchange
    to_holder: str = in_transaction.holder
    spot_price: str = _resolve_fields(
        f"out_transaction.{Keyword.SPOT_PRICE.value}",
        f"in_transaction.{Keyword.SPOT_PRICE.value}",
        out_transaction.spot_price,
        in_transaction.spot_price,
        out_transaction,
        in_transaction,
        disallow_two_unknown=False,
    )
    # If one of the transaction has spot_price from web, and the other doesn't take the non-web one
    if out_transaction.is_spot_price_from_web:
        spot_price = _resolve_fields(
            f"in_transaction.{Keyword.SPOT_PRICE.value}",
            f"out_transaction.{Keyword.SPOT_PRICE.value}",
            in_transaction.spot_price,
            out_transaction.spot_price,
            in_transaction,
            out_transaction,
            disallow_two_unknown=False,
            if_conflict_override_second_parameter=True,
        )
    elif in_transaction.is_spot_price_from_web:
        spot_price = _resolve_fields(
            f"out_transaction.{Keyword.SPOT_PRICE.value}",
            f"in_transaction.{Keyword.SPOT_PRICE.value}",
            out_transaction.spot_price,
            in_transaction.spot_price,
            out_transaction,
            in_transaction,
            disallow_two_unknown=False,
            if_conflict_override_second_parameter=True,
        )
    else:
        spot_price = _resolve_fields(
            f"out_transaction.{Keyword.SPOT_PRICE.value}",
            f"in_transaction.{Keyword.SPOT_PRICE.value}",
            out_transaction.spot_price,
            in_transaction.spot_price,
            out_transaction,
            in_transaction,
            disallow_two_unknown=False,
            if_conflict_override_second_parameter=False,
        )

    crypto_sent: str = str(RP2Decimal(out_transaction.crypto_out_no_fee) + RP2Decimal(out_transaction.crypto_fee))
    crypto_received: str = in_transaction.crypto_in

    notes = f"{notes}; " if notes else ""
    notes += f"{in_transaction.notes}; " if in_transaction.notes else ""
    notes += f"{out_transaction.notes}; " if out_transaction.notes else ""

    transaction = IntraTransaction(
        plugin=__RESOLVER,
        unique_id=in_transaction.unique_id,
        raw_data=f"{in_transaction.raw_data}\n{out_transaction.raw_data}",
        asset=in_transaction.asset,
        timestamp=str(timestamp),
        from_exchange=from_exchange,
        from_holder=from_holder,
        to_exchange=to_exchange,
        to_holder=to_holder,
        spot_price=spot_price,
        crypto_sent=crypto_sent,
        crypto_received=crypto_received,
        notes=notes,
    )

    return transaction



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from inspect import signature
from typing import Callable, Dict, List, NamedTuple, Optional, Union

from dateutil.parser import parse
from rp2.configuration import to_string

from dali.dali_configuration import Keyword, is_internal_field, is_unknown


class StringAndDatetime(NamedTuple):
    string: str
    value: datetime


class AbstractTransaction:

    @classmethod
    def _validate_string_field(cls, name: str, value: str, raw_data: str, disallow_empty: bool, disallow_unknown: bool) -> str:
        if not isinstance(name, str):
            raise Exception(f"Internal error: parameter name is not a string: {name}")
        if not isinstance(value, str):
            raise Exception(f"Internal error: {name} is not a string: {value}\n{raw_data}")
        if disallow_empty and len(value) == 0:
            raise Exception(f"Internal error: {name} is empty: {raw_data}")
        if disallow_unknown and is_unknown(value):
            raise Exception(f"Internal error: {name} is unknown: {raw_data}")
        return value.strip()

    @classmethod
    def _validate_numeric_field(cls, name: str, value: str, raw_data: str, disallow_empty: bool, disallow_unknown: bool) -> str:
        value = cls._validate_string_field(name, value, raw_data, disallow_empty, disallow_unknown)
        if is_unknown(value):
            return value
        if not value:
            return value
        try:
            float(value)
        except ValueError as exc:
            raise Exception(f"Internal error parsing {name} as number: {value}\n{raw_data}\n{str(exc)}") from exc
        return value

    @classmethod
    def _validate_timestamp_field(cls, name: str, value: str, raw_data: str) -> StringAndDatetime:
        value = cls._validate_string_field(name, value, raw_data, disallow_empty=True, disallow_unknown=True)
        try:
            result: datetime = parse(value)
        except Exception as exc:
            raise Exception(f"Internal error parsing {name} as datetime: {value}\n{raw_data}\n{str(exc)}") from exc
        if result.tzinfo is None:
            raise Exception(f"Internal error: {name} has no timezone info: {value}\n{raw_data}")
        if result.microsecond == 0:
            return StringAndDatetime(result.strftime("%Y-%m-%d %H:%M:%S %z"), result)
        return StringAndDatetime(result.strftime("%Y-%m-%d %H:%M:%S.%f %z"), result)

    @classmethod
    def _validate_optional_string_field(cls, name: str, value: Optional[str], raw_data: str, disallow_empty: bool, disallow_unknown: bool) -> Optional[str]:
        if not value:
            return None
        return cls._validate_string_field(name, value, raw_data, disallow_empty, disallow_unknown)

    @classmethod
    def _validate_optional_numeric_field(cls, name: str, value: Optional[str], raw_data: str, disallow_empty: bool, disallow_unknown: bool) -> Optional[str]:
        if not value:
            return None
        return cls._validate_numeric_field(name, value, raw_data, disallow_empty, disallow_unknown)

    # Unique_id is used by transaction_resolver: it must contain hash for IntraTransactions or other account-specific id
    # that identifies the transaction uniquely. Some exchanges don't report hash information, so allow unknown (e.g. BlockFi CSV)
    def __init__(
        self,
        plugin: str,
        unique_id: str,
        raw_data: str,
        timestamp: str,
        asset: str,
        notes: Optional[str] = None,
        is_spot_price_from_web: Optional[bool] = None,
    ) -> None:
        self.__plugin: str = self._validate_string_field(Keyword.PLUGIN.value, plugin, raw_data, disallow_empty=True, disallow_unknown=True)
        self.__unique_id: str = self._validate_string_field(Keyword.UNIQUE_ID.value, unique_id, raw_data, disallow_empty=True, disallow_unknown=False)
        if unique_id.startswith("0x"):
            self.__unique_id = unique_id[len("0x") :]
        self.__raw_data: str = self._validate_string_field(Keyword.RAW_DATA.value, raw_data, raw_data, disallow_empty=True, disallow_unknown=True)
        self.__timestamp: str
        self.__timestamp_value: datetime
        (self.__timestamp, self.__timestamp_value) = self._validate_timestamp_field(Keyword.TIMESTAMP.value, timestamp, raw_data)
        self.__asset: str = self._validate_string_field(Keyword.ASSET.value, asset, raw_data, disallow_empty=True, disallow_unknown=True)
        self.__notes: Optional[str] = self._validate_optional_string_field(Keyword.NOTES.value, notes, raw_data, disallow_empty=False, disallow_unknown=True)
        if is_spot_price_from_web and not isinstance(is_spot_price_from_web, bool):
            raise Exception(f"Internal error: {Keyword.IS_SPOT_PRICE_FROM_WEB.value} is not boolean: {is_spot_price_from_web}")
        self.__is_spot_price_from_web: bool = is_spot_price_from_web if is_spot_price_from_web else False

    def to_string(self, indent: int = 0, repr_format: bool = True, extra_data: Optional[List[str]] = None) -> str:
        class_specific_data: List[str] = []
        stringify: Callable[[object], str] = repr
        if not repr_format:
            stringify = str

        if repr_format:
            class_specific_data.append(f"{type(self).__name__}({Keyword.PLUGIN.value}={repr(self.plugin)}")
        else:
            class_specific_data.append(f"{type(self).__name__}:")
            class_specific_data.append(f"{Keyword.PLUGIN.value}={str(self.plugin)}")

        class_specific_data.append(f"{Keyword.UNIQUE_ID.value}={stringify(self.unique_id)}")
        class_specific_data.append(f"{Keyword.RAW_DATA.value}={stringify(self.raw_data)}")
        class_specific_data.append(f"{Keyword.TIMESTAMP.value}={stringify(self.timestamp)}")
        class_specific_data.append(f"{Keyword.ASSET.value}={stringify(self.asset)}")

        if extra_data:
            class_specific_data.extend(extra_data)

        return to_string(indent=indent, repr_format=repr_format, data=class_specific_data)

    def __str__(self) -> str:
        return self.to_string(indent=0, repr_format=False)

    def __repr__(self) -> str:
        return self.to_string(indent=0, repr_format=True)

    def __eq__(self, other: object) -> bool:
        if not other:
            return False
        if not isinstance(other, AbstractTransaction):
            raise Exception(f"Internal error: operand has non-AbstractTransaction value {repr(other)}")
        result: bool = self.unique_id == other.unique_id and self.plugin == self.plugin and self.asset == self.asset
        return result

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((self.unique_id, self.plugin, self.asset))

    # Build a dictionary of constructor initialization parameters. Return true if any of them have UNKNOWN value
    def _setup_constructor_parameter_dictionary(self, parameter_dictionary: Dict[str, Union[str, bool, Optional[str], Optional[bool]]]) -> bool:
        result: bool = False
        for parameter in signature(self.__class__).parameters:
            value: str = getattr(self, parameter)
            parameter_dictionary[parameter] = value
            if is_internal_field(parameter) or parameter == Keyword.UNIQUE_ID.value:
                continue
            if is_unknown(value):
                result = True
        return result

    @property
    def plugin(self) -> str:
        return self.__plugin

    @property
    def unique_id(self) -> str:
        return self.__unique_id

    @property
    def raw_data(self) -> str:
        return self.__raw_data

    @property
    def timestamp(self) -> str:
        return self.__timestamp

    @property
    def timestamp_value(self) -> datetime:
        return self.__timestamp_value

    @property
    def asset(self) -> str:
        return self.__asset

    @property
    def notes(self) -> Optional[str]:
        return self.__notes

    @property
    def is_spot_price_from_web(self) -> bool:
        return self.__is_spot_price_from_web

    @property
    def is_unresolved(self) -> bool:
        raise NotImplementedError("Abstract method: it must be implemented in subclasses")

    @property
    def constructor_parameter_dictionary(self) -> Dict[str, Union[str, bool, Optional[str], Optional[bool]]]:
        raise NotImplementedError("Abstract method: it must be implemented in subclasses")



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Dict, List, Optional, Union

from dali.abstract_transaction import AbstractTransaction
from dali.dali_configuration import Keyword, is_transaction_type_valid


class InTransaction(AbstractTransaction):
    @classmethod
    def _validate_transaction_type_field(cls, name: str, value: str, raw_data: str) -> str:
        value = cls._validate_string_field(name, value, raw_data, disallow_empty=True, disallow_unknown=True)
        Keyword.type_check_from_string(value)
        if not is_transaction_type_valid(Keyword.IN.value, value):
            raise Exception(f"Invalid transaction type {value} for {cls.__name__}")
        return value

    def __init__(
        self,
        plugin: str,
        unique_id: str,
        raw_data: str,
        timestamp: str,
        asset: str,
        exchange: str,
        holder: str,
        transaction_type: str,
        spot_price: str,
        crypto_in: str,
        fiat_fee: str,
        fiat_in_no_fee: Optional[str] = None,
        fiat_in_with_fee: Optional[str] = None,
        notes: Optional[str] = None,
        is_spot_price_from_web: Optional[bool] = None,
    ) -> None:
        super().__init__(plugin, unique_id, raw_data, timestamp, asset, notes, is_spot_price_from_web)

        self.__exchange: str = self._validate_string_field(Keyword.EXCHANGE.value, exchange, raw_data, disallow_empty=True, disallow_unknown=True)
        self.__holder: str = self._validate_string_field(Keyword.HOLDER.value, holder, raw_data, disallow_empty=True, disallow_unknown=True)
        self.__transaction_type: str = self._validate_transaction_type_field(Keyword.TRANSACTION_TYPE.value, transaction_type, raw_data)
        self.__spot_price: str = self._validate_numeric_field(Keyword.SPOT_PRICE.value, spot_price, raw_data, disallow_empty=True, disallow_unknown=False)
        self.__crypto_in: str = self._validate_numeric_field(Keyword.CRYPTO_IN.value, crypto_in, raw_data, disallow_empty=True, disallow_unknown=True)
        self.__fiat_fee: str = self._validate_numeric_field(Keyword.FIAT_FEE.value, fiat_fee, raw_data, disallow_empty=True, disallow_unknown=True)
        self.__fiat_in_no_fee: Optional[str] = self._validate_optional_numeric_field(
            Keyword.FIAT_IN_NO_FEE.value, fiat_in_no_fee, raw_data, disallow_empty=False, disallow_unknown=True
        )
        self.__fiat_in_with_fee: Optional[str] = self._validate_optional_numeric_field(
            Keyword.FIAT_IN_WITH_FEE.value, fiat_in_with_fee, raw_data, disallow_empty=False, disallow_unknown=True
        )
        self.__constructor_parameter_dictionary: Dict[str, Union[str, bool, Optional[str], Optional[bool]]] = {}
        self.__is_unresolved: bool = self._setup_constructor_parameter_dictionary(self.__constructor_parameter_dictionary)

    def to_string(self, indent: int = 0, repr_format: bool = True, extra_data: Optional[List[str]] = None) -> str:

        class_specific_data: List[str] = []
        stringify: Callable[[object], str] = repr
        if not repr_format:
            stringify = str
        class_specific_data = [
            f"{Keyword.EXCHANGE.value}={stringify(self.exchange)}",
            f"{Keyword.HOLDER.value}={stringify(self.holder)}",
            f"{Keyword.TRANSACTION_TYPE.value}={stringify(self.transaction_type)}",
            f"{Keyword.SPOT_PRICE.value}={self.spot_price}",
            f"{Keyword.CRYPTO_IN.value}={self.crypto_in}",
            f"{Keyword.FIAT_FEE.value}={self.fiat_fee}",
            f"{Keyword.FIAT_IN_NO_FEE.value}={self.fiat_in_no_fee}",
            f"{Keyword.FIAT_IN_WITH_FEE.value}={self.fiat_in_with_fee}",
            f"{Keyword.NOTES.value}={self.notes}",
        ]
        if extra_data:
            class_specific_data.extend(extra_data)

        return super().to_string(indent=indent, repr_format=repr_format, extra_data=class_specific_data)

    @property
    def exchange(self) -> str:
        return self.__exchange

    @property
    def holder(self) -> str:
        return self.__holder

    @property
    def transaction_type(self) -> str:
        return self.__transaction_type

    @property
    def spot_price(self) -> str:
        return self.__spot_price

    @property
    def crypto_in(self) -> str:
        return self.__crypto_in

    @property
    def fiat_in_no_fee(self) -> Optional[str]:
        return self.__fiat_in_no_fee

    @property
    def fiat_in_with_fee(self) -> Optional[str]:
        return self.__fiat_in_with_fee

    @property
    def fiat_fee(self) -> str:
        return self.__fiat_fee

    @property
    def is_unresolved(self) -> bool:
        return self.__is_unresolved

    @property
    def constructor_parameter_dictionary(self) -> Dict[str, Union[str, bool, Optional[str], Optional[bool]]]:
        return self.__constructor_parameter_dictionary



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# CSV Format: currency, amount, type, timestamp

import logging
from csv import reader
from typing import List, Optional

from rp2.logger import create_logger
from rp2.rp2_decimal import RP2Decimal

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.dali_configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction

_CRYPTO_TRANSFER: str = "Crypto Transfer"
_INTEREST_PAYMENT: str = "Interest Payment"
_WITHDRAWAL: str = "Withdrawal"
_WITHDRAWAL_FEE: str = "Withdrawal Fee"


class InputPlugin(AbstractInputPlugin):

    __BLOCKFI: str = "BlockFi"

    __CURRENCY_INDEX: int = 0
    __AMOUNT_INDEX: int = 1
    __TYPE_INDEX: int = 2
    __TIMESTAMP_INDEX: int = 3
    #    __SPOT_PRICE_INDEX: int = 4 # It used to be present in the CSV, but sometime in 2021 BlockFi removed this field.

    __DELIMITER = ","

    def __init__(
        self,
        account_holder: str,
        csv_file: str,
    ) -> None:

        super().__init__(account_holder)
        self.__csv_file: str = csv_file

        self.__logger: logging.Logger = create_logger(f"{self.__BLOCKFI}/{self.account_holder}")

    def load(self) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        last_withdrawal_fee: Optional[RP2Decimal] = None
        with open(self.__csv_file, mode="r", encoding="utf-8") as csv_file:
            lines = reader(csv_file)
            header_found: bool = False
            for line in lines:
                raw_data: str = self.__DELIMITER.join(line)
                if not header_found:
                    # Skip header line
                    header_found = True
                    self.__logger.debug("Header: %s", raw_data)
                    continue
                self.__logger.debug("Transaction: %s", raw_data)

                if last_withdrawal_fee is not None and line[self.__TYPE_INDEX] != _WITHDRAWAL:
                    raise Exception(f"Internal error: withdrawal fee {last_withdrawal_fee} is not followed by withdrawal")

                if line[self.__TYPE_INDEX] == _INTEREST_PAYMENT:
                    last_withdrawal_fee = None
                    result.append(
                        InTransaction(
                            self.__BLOCKFI,
                            Keyword.UNKNOWN.value,
                            raw_data,
                            f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                            line[self.__CURRENCY_INDEX],
                            self.__BLOCKFI,
                            self.account_holder,
                            "Interest",
                            Keyword.UNKNOWN.value,
                            line[self.__AMOUNT_INDEX],
                            "0",
                        )
                    )
                elif line[self.__TYPE_INDEX] == _CRYPTO_TRANSFER:
                    last_withdrawal_fee = None
                    result.append(
                        IntraTransaction(
                            self.__BLOCKFI,
                            Keyword.UNKNOWN.value,
                            raw_data,
                            f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                            line[self.__CURRENCY_INDEX],
                            Keyword.UNKNOWN.value,
                            Keyword.UNKNOWN.value,
                            self.__BLOCKFI,
                            self.account_holder,
                            Keyword.UNKNOWN.value,
                            Keyword.UNKNOWN.value,
                            line[self.__AMOUNT_INDEX],
                        )
                    )
                elif line[self.__TYPE_INDEX] == _WITHDRAWAL:
                    amount: RP2Decimal = RP2Decimal(line[self.__AMOUNT_INDEX])
                    amount = -amount  # type: ignore
                    if last_withdrawal_fee is not None:
                        amount += last_withdrawal_fee
                    last_withdrawal_fee = None
                    result.append(
                        IntraTransaction(
                            self.__BLOCKFI,
                            Keyword.UNKNOWN.value,
                            raw_data,
                            f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                            line[self.__CURRENCY_INDEX],
                            self.__BLOCKFI,
                            self.account_holder,
                            Keyword.UNKNOWN.value,
                            Keyword.UNKNOWN.value,
                            Keyword.UNKNOWN.value,
                            str(amount),
                            Keyword.UNKNOWN.value,
                        )
                    )
                elif line[self.__TYPE_INDEX] == _WITHDRAWAL_FEE:
                    last_withdrawal_fee = RP2Decimal(line[self.__AMOUNT_INDEX])
                    last_withdrawal_fee = -last_withdrawal_fee  # type: ignore
                else:
                    self.__logger.debug("Unsupported transaction type (skipping): %s", raw_data)
        return result



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from pathlib import Path
from typing import Any, Dict, List, Set

from dali.abstract_transaction import AbstractTransaction
from dali.dali_configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction

_ASSETS: str = "assets"
_EXCHANGES: str = "exchanges"
_HOLDERS: str = "holders"


def generate_config_file(
    output_dir_path: str,
    output_file_prefix: str,
    output_file_name: str,
    transactions: List[AbstractTransaction],
    global_configuration: Dict[str, Any],
) -> Any:

    if not isinstance(output_dir_path, str):
        raise Exception(f"Internal error: parameter output_dir_path is not of type string: {repr(output_dir_path)}")
    if not isinstance(output_file_prefix, str):
        raise Exception(f"Internal error: parameter output_file_prefix is not of type string: {repr(output_file_prefix)}")
    if not isinstance(output_file_name, str):
        raise Exception(f"Internal error: parameter output_file_name is not of type string: {repr(output_file_name)}")
    if not isinstance(transactions, List):
        raise Exception(f"Internal error: parameter transactions is not of type List: {repr(transactions)}")

    output_file_path: Path = Path(output_dir_path) / Path(f"{output_file_prefix}{output_file_name}")
    if Path(output_file_path).exists():
        output_file_path.unlink()

    json_config: Dict[str, Any] = {
        Keyword.IN_HEADER.value: {},
        Keyword.OUT_HEADER.value: {},
        Keyword.INTRA_HEADER.value: {},
        _ASSETS: [],
        _EXCHANGES: [],
        _HOLDERS: [],
    }

    assets: Set[str] = set()
    holders: Set[str] = set()
    exchanges: Set[str] = set()
    for transaction in transactions:
        if isinstance(transaction, InTransaction):
            holders.add(transaction.holder)
            exchanges.add(transaction.exchange)
        elif isinstance(transaction, OutTransaction):
            holders.add(transaction.holder)
            exchanges.add(transaction.exchange)
        elif isinstance(transaction, IntraTransaction):
            holders.add(transaction.from_holder)
            holders.add(transaction.to_holder)
            exchanges.add(transaction.from_exchange)
            exchanges.add(transaction.to_exchange)
        else:
            raise Exception(f"Internal error: transaction is not a subclass of AbstractTransaction: {transaction}")
        assets.add(transaction.asset)

    for section_name in [Keyword.IN_HEADER, Keyword.OUT_HEADER, Keyword.INTRA_HEADER]:
        json_config[section_name.value] = global_configuration[section_name.value]

    if Keyword.UNKNOWN.value in assets:
        assets.remove(Keyword.UNKNOWN.value)
    if Keyword.UNKNOWN.value in holders:
        holders.remove(Keyword.UNKNOWN.value)
    if Keyword.UNKNOWN.value in exchanges:
        exchanges.remove(Keyword.UNKNOWN.value)
    json_config[_ASSETS] = list(assets)
    json_config[_HOLDERS] = list(holders)
    json_config[_EXCHANGES] = list(exchanges)

    with open(str(output_file_path), "w", encoding="utf-8") as output_file:
        json.dump(json_config, output_file, indent=4)



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=line-too-long

# IN CSV Format: unique_id, timestamp, asset, exchange, holder, transaction_type, spot_price, crypto_in, fiat_fee, fiat_in_no_fee, fiat_in_with_fee, notes
# OUT CSV Format: unique_id, timestamp, asset, exchange, holder, transaction_type, spot_price, crypto_out_no_fee, crypto_fee, crypto_out_with_fee, fiat_in_no_fee, fiat_fee, notes
# INTRA CSV Format:unique_id, timestamp, asset, from_exchange, from_holder, to_exchange, to_holder, spot_price, crypto_sent, crypto_received, notes

# pylint: enable=line-too-long

import logging
from csv import reader
from typing import List

from rp2.logger import create_logger

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.dali_configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction


class InputPlugin(AbstractInputPlugin):

    __MANUAL: str = "Manual"

    __IN_UNIQUE_ID_INDEX: int = 0
    __IN_TIMESTAMP_INDEX: int = 1
    __IN_ASSET_INDEX: int = 2
    __IN_EXCHANGE_INDEX: int = 3
    __IN_HOLDER_INDEX: int = 4
    __IN_TYPE_INDEX: int = 5
    __IN_SPOT_PRICE_INDEX: int = 6
    __IN_CRYPTO_IN_INDEX: int = 7
    __IN_FIAT_FEE_INDEX: int = 8
    __IN_FIAT_IN_NO_FEE_INDEX: int = 9
    __IN_FIAT_IN_WITH_FEE_INDEX: int = 10
    __IN_NOTES_INDEX: int = 11

    __OUT_UNIQUE_ID_INDEX: int = 0
    __OUT_TIMESTAMP_INDEX: int = 1
    __OUT_ASSET_INDEX: int = 2
    __OUT_EXCHANGE_INDEX: int = 3
    __OUT_HOLDER_INDEX: int = 4
    __OUT_TYPE_INDEX: int = 5
    __OUT_SPOT_PRICE_INDEX: int = 6
    __OUT_CRYPTO_OUT_NO_FEE_INDEX: int = 7
    __OUT_CRYPTO_FEE_INDEX: int = 8
    __OUT_CRYPTO_OUT_WITH_FEE_INDEX: int = 9
    __OUT_FIAT_OUT_NO_FEE_INDEX: int = 10
    __OUT_FIAT_FEE_INDEX: int = 11
    __OUT_NOTES_INDEX: int = 12

    __INTRA_UNIQUE_ID_INDEX: int = 0
    __INTRA_TIMESTAMP_INDEX: int = 1
    __INTRA_ASSET_INDEX: int = 2
    __INTRA_FROM_EXCHANGE_INDEX: int = 3
    __INTRA_FROM_HOLDER_INDEX: int = 4
    __INTRA_TO_EXCHANGE_INDEX: int = 5
    __INTRA_TO_HOLDER_INDEX: int = 6
    __INTRA_SPOT_PRICE_INDEX: int = 7
    __INTRA_CRYPTO_SENT_INDEX: int = 8
    __INTRA_CRYPTO_RECEIVED_INDEX: int = 9
    __INTRA_NOTES_INDEX: int = 10

    def __init__(
        self,
        in_csv_file: str,
        out_csv_file: str,
        intra_csv_file: str,
    ) -> None:

        super().__init__("")

        self.__in_csv_file: str = in_csv_file
        self.__out_csv_file: str = out_csv_file
        self.__intra_csv_file: str = intra_csv_file

        self.__logger: logging.Logger = create_logger(self.__MANUAL)

    def load(self) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        self._load_in_file(result)
        self._load_out_file(result)
        self._load_intra_file(result)

        return result

    def _load_in_file(self, transactions: List[AbstractTransaction]) -> None:

        if not self.__in_csv_file:
            return

        with open(self.__in_csv_file, mode="r", encoding="utf-8") as csv_file:
            lines = reader(csv_file)
            header_found: bool = False
            for line in lines:
                raw_data: str = ",".join(line).strip()
                if not header_found:
                    # Skip header line
                    header_found = True
                    self.__logger.debug("Header: %s", ";".join(line))
                    continue

                if raw_data.startswith("," * self.__IN_NOTES_INDEX):
                    # Skip empty lines
                    continue

                unique_id: str = line[self.__IN_UNIQUE_ID_INDEX]
                if not unique_id:
                    unique_id = Keyword.UNKNOWN.value
                self.__logger.debug("Transaction: %s", ",".join(line))
                transactions.append(
                    InTransaction(
                        self.__MANUAL,
                        unique_id,
                        raw_data,
                        line[self.__IN_TIMESTAMP_INDEX],
                        line[self.__IN_ASSET_INDEX],
                        line[self.__IN_EXCHANGE_INDEX],
                        line[self.__IN_HOLDER_INDEX],
                        line[self.__IN_TYPE_INDEX],
                        line[self.__IN_SPOT_PRICE_INDEX],
                        line[self.__IN_CRYPTO_IN_INDEX],
                        line[self.__IN_FIAT_FEE_INDEX],
                        line[self.__IN_FIAT_IN_NO_FEE_INDEX],
                        line[self.__IN_FIAT_IN_WITH_FEE_INDEX],
                        line[self.__IN_NOTES_INDEX],
                    )
                )

    def _load_out_file(self, transactions: List[AbstractTransaction]) -> None:

        if not self.__out_csv_file:
            return

        with open(self.__out_csv_file, mode="r", encoding="utf-8") as csv_file:
            lines = reader(csv_file)
            header_found: bool = False
            for line in lines:
                raw_data: str = ",".join(line).strip()
                if not header_found:
                    # Skip header line
                    header_found = True
                    self.__logger.debug("Header: %s", ";".join(line))
                    continue

                if raw_data.startswith("," * self.__OUT_NOTES_INDEX):
                    # Skip empty lines
                    continue

                unique_id: str = line[self.__OUT_UNIQUE_ID_INDEX]
                if not unique_id:
                    unique_id = Keyword.UNKNOWN.value
                self.__logger.debug("Transaction: %s", ",".join(line))
                transactions.append(
                    OutTransaction(
                        self.__MANUAL,
                        unique_id,
                        raw_data,
                        line[self.__OUT_TIMESTAMP_INDEX],
                        line[self.__OUT_ASSET_INDEX],
                        line[self.__OUT_EXCHANGE_INDEX],
                        line[self.__OUT_HOLDER_INDEX],
                        line[self.__OUT_TYPE_INDEX],
                        line[self.__OUT_SPOT_PRICE_INDEX],
                        line[self.__OUT_CRYPTO_OUT_NO_FEE_INDEX],
                        line[self.__OUT_CRYPTO_FEE_INDEX],
                        line[self.__OUT_CRYPTO_OUT_WITH_FEE_INDEX],
                        line[self.__OUT_FIAT_OUT_NO_FEE_INDEX],
                        line[self.__OUT_FIAT_FEE_INDEX],
                        line[self.__OUT_NOTES_INDEX],
                    )
                )

    def _load_intra_file(self, transactions: List[AbstractTransaction]) -> None:

        if not self.__intra_csv_file:
            return

        with open(self.__intra_csv_file, mode="r", encoding="utf-8") as csv_file:
            lines = reader(csv_file)
            header_found: bool = False
            for line in lines:
                raw_data: str = ",".join(line).strip()
                if not header_found:
                    # Skip header line
                    header_found = True
                    self.__logger.debug("Header: %s", ";".join(line))
                    continue

                if raw_data.startswith("," * self.__INTRA_NOTES_INDEX):
                    # Skip empty lines
                    continue

                self.__logger.debug("Transaction: %s", ",".join(line))
                from_exchange: str = line[self.__INTRA_FROM_EXCHANGE_INDEX]
                from_holder: str = line[self.__INTRA_FROM_HOLDER_INDEX]
                to_exchange: str = line[self.__INTRA_TO_EXCHANGE_INDEX]
                to_holder: str = line[self.__INTRA_TO_HOLDER_INDEX]
                crypto_sent: str = line[self.__INTRA_CRYPTO_SENT_INDEX]
                crypto_received: str = line[self.__INTRA_CRYPTO_RECEIVED_INDEX]
                transactions.append(
                    IntraTransaction(
                        self.__MANUAL,
                        line[self.__INTRA_UNIQUE_ID_INDEX],
                        raw_data,
                        line[self.__INTRA_TIMESTAMP_INDEX],
                        line[self.__INTRA_ASSET_INDEX],
                        from_exchange if from_exchange else Keyword.UNKNOWN.value,
                        from_holder if from_holder else Keyword.UNKNOWN.value,
                        to_exchange if to_exchange else Keyword.UNKNOWN.value,
                        to_holder if to_holder else Keyword.UNKNOWN.value,
                        line[self.__INTRA_SPOT_PRICE_INDEX],
                        crypto_sent if crypto_sent else Keyword.UNKNOWN.value,
                        crypto_received if crypto_received else Keyword.UNKNOWN.value,
                        line[self.__INTRA_NOTES_INDEX],
                    )
                )
